"""
Utility functions to access the entry points registry
"""

# Standard library modules.
from operator import attrgetter

# Third party modules.
import pkg_resources

# Local modules.

# Globals and constants variables.

def iter_entry_points(group, name=None):
    entry_points = list(pkg_resources.iter_entry_points(group, name)) #@UndefinedVariable
    entry_points.sort(key=attrgetter('name'))

    for entry_point in entry_points:
        yield entry_point.name, entry_point.load(require=False)

def iter_condition_widget_classes(name=None):
    return iter_entry_points('pyhmsa.gui.spec.condition', name)

def iter_condition_widgets(name=None, *args, **kwargs):
    for name, clasz in iter_condition_widget_classes(name):
        yield name, clasz(*args, **kwargs)

def iter_datum_widget_classes(name=None):
    return iter_entry_points('pyhmsa.gui.spec.datum', name)

def iter_datum_widgets(name=None, *args, **kwargs):
    for name, clasz in iter_datum_widget_classes(name):
        yield name, clasz(*args, **kwargs)

def iter_importer_classes(name=None):
    for name, clasz in iter_entry_points('pyhmsa.fileformat.importer', name):
        for ext in clasz.SUPPORTED_EXTENSIONS:
            yield name, ext, clasz

def iter_importers(name=None, *args, **kwargs):
    for name, ext, clasz in iter_importer_classes(name):
        yield name, ext, clasz(*args, **kwargs)

def iter_exporter_classes(name=None):
    return iter_entry_points('pyhmsa.fileformat.exporter', name)

def iter_exporters(name=None, *args, **kwargs):
    for name, clasz in iter_exporter_classes(name):
        yield name, clasz(*args, **kwargs)

def iter_preferences_widget_classes(name=None):
    return iter_entry_points('pyhmsa.viewer.preferences', name)


"""
Extension of matplotlib backend
"""

# Standard library modules.
import os
from operator import itemgetter

# Third party modules.
import six

import numpy as np

from PySide.QtCore import Qt, Signal
from PySide.QtGui import \
    (QAction, QDialog, QComboBox, QVBoxLayout, QDialogButtonBox, QCheckBox,
     QFormLayout, QSlider, QLabel, QFileDialog, QSpinBox, QHBoxLayout,
     QMessageBox, QWidget, QButtonGroup, QRadioButton)

import matplotlib
import matplotlib.cm as cm
import matplotlib.colors as colors
from matplotlib.backends.backend_qt5 import NavigationToolbar2QT as _NavigationToolbar2QT

# Local modules.
from pyhmsa.gui.util.icon import getIcon
from pyhmsa.gui.util.layout import horizontal
from pyhmsa.gui.util.widget import ColorDialogButton
from pyhmsa.gui.util.mpl.scalebar import ScaleBar
from pyhmsa.gui.util.mpl.color import \
    convert_color_mpl_to_qt, convert_color_qt_to_mpl

# Globals and constants variables.

class _SaveDialog(QFileDialog):

    def __init__(self, parent):
        QFileDialog.__init__(self, parent)
        self.setFileMode(QFileDialog.FileMode.AnyFile)
        self.setAcceptMode(QFileDialog.AcceptMode.AcceptSave)

        # Widgets
        self._chk_tight = QCheckBox('Tight layout')

        self._txt_dpi = QSpinBox()
        self._txt_dpi.setRange(1, 10000)
        self._txt_dpi.setSingleStep(50)
        self._txt_dpi.setSuffix('dpi')
        self._txt_dpi.setValue(100)

        # Layouts
        layout = self.layout()

        lyt_extras = QHBoxLayout()
        lyt_extras.addWidget(QLabel('Extra options'))
        lyt_extras.addWidget(self._chk_tight)
        lyt_extras.addWidget(QLabel('Resolution'))
        lyt_extras.addWidget(self._txt_dpi)
        layout.addLayout(lyt_extras, layout.rowCount(), 0, 1, layout.columnCount())

        self.setLayout(layout)

    def tightLayout(self):
        return self._chk_tight.isChecked()

    def setTightLayout(self, tight):
        self._chk_tight.setChecked(tight)

    def dpi(self):
        return self._txt_dpi.value()

    def setDpi(self, dpi):
        self._txt_dpi.setValue(dpi)

class NavigationToolbarQT(_NavigationToolbar2QT):

    def _get_save_name_filters(self):
        filters = []
        selected_filter = None
        filetypes = self.canvas.get_supported_filetypes_grouped()
        sorted_filetypes = list(six.iteritems(filetypes))
        sorted_filetypes.sort()
        default_filetype = self.canvas.get_default_filetype()

        for name, exts in sorted_filetypes:
            exts_list = " ".join(['*.%s' % ext for ext in exts])
            filter_ = '%s (%s)' % (name, exts_list)
            if default_filetype in exts:
                selected_filter = filter_
            filters.append(filter_)

        return filters, selected_filter

    def save_figure(self, *args):
        dialog = _SaveDialog(self.parent)
        dialog.setWindowTitle("Choose a filename to save to")
        dialog.setDpi(matplotlib.rcParams.get('savefig.dpi',
                                              self.canvas.figure.dpi))
        dialog.setTightLayout(matplotlib.rcParams.get('savefig.bbox', None) == 'tight')

        filters, selected_filter = self._get_save_name_filters()
        dialog.setNameFilters(filters)
        dialog.selectNameFilter(selected_filter)

        startpath = matplotlib.rcParams.get('savefig.directory', '')
        startpath = os.path.expanduser(startpath)
        start = os.path.join(startpath, self.canvas.get_default_filename())
        dialog.selectFile(start)

        if not dialog.exec_():
            return

        fnames = dialog.selectedFiles()
        if len(fnames) != 1:
            return
        fname = fnames[0]

        dpi = dialog.dpi()
        bbox_inches = 'tight' if dialog.tightLayout() else None

        # Store default values
        if startpath == '':
            # explicitly missing key or empty str signals to use cwd
            matplotlib.rcParams['savefig.directory'] = startpath
        else:
            # save dir for next time
            savefig_dir = os.path.dirname(six.text_type(fname))
            matplotlib.rcParams['savefig.directory'] = savefig_dir

        matplotlib.rcParams['savefig.dpi'] = dpi
        matplotlib.rcParams['savefig.bbox'] = bbox_inches

        # Save
        try:
            self.canvas.print_figure(six.text_type(fname),
                                 dpi=dpi, bbox_inches=bbox_inches)
        except Exception as e:
            QMessageBox.critical(
                self, "Error saving file", str(e),
                QMessageBox.Ok, QMessageBox.NoButton)

class NavigationToolbarSnapMixin(object):

    def __init__(self):
        self._snap_cross = {}

    def _clear_snap(self):
        for axes, (crossh, crossv) in self._snap_cross.items():
            axes.lines.remove(crossh)
            axes.lines.remove(crossv)
        self._snap_cross.clear()
        self.draw()

    def pan(self, *args):
        self._clear_snap()
        super().pan(*args)

    def zoom(self, *args):
        print('xoom')
        self._clear_snap()
        super().zoom(*args)

    def snap(self, *args):
        if self._active == 'SNAP':
            self._active = None
        else:
            self._active = 'SNAP'

        if self._idPress is not None:
            self._idPress = self.canvas.mpl_disconnect(self._idPress)

        if self._idRelease is not None:
            self._idRelease = self.canvas.mpl_disconnect(self._idRelease)

        self._clear_snap()

        if self._active:
            self.mode = 'snap'
            self.canvas.widgetlock(self)

            for axes in self.canvas.figure.get_axes():
                xmin, _ = axes.get_xlim()
                ymin, _ = axes.get_ylim()
                color = matplotlib.rcParams.get('snap.color', 'r')
                lw = matplotlib.rcParams.get('snap.linewidth', 2)
                crossh = axes.axhline(ymin, color=color, lw=lw)
                crossv = axes.axvline(xmin, color=color, lw=lw)
                crossh.set_visible(False)
                crossv.set_visible(False)
                self._snap_cross[axes] = (crossh, crossv)
        else:
            self.mode = ''
            self.canvas.widgetlock.release(self)

        for a in self.canvas.figure.get_axes():
            a.set_navigate_mode(self._active)

        self.set_message(self.mode)

    def mouse_move(self, event):
        if self.mode != 'snap':
            super().mouse_move(event)
            return

        for axes, (crossh, crossv) in self._snap_cross.items():
            crossh.set_visible(event.inaxes == axes)
            crossv.set_visible(event.inaxes == axes)

        if event.inaxes is None:
            super().mouse_move(event)
            self.draw()
            return

        x = event.xdata
        y = event.ydata

        xlines = []
        ylines = []
        for line in event.inaxes.lines:
            if line in self._snap_cross[event.inaxes]: continue
            xs = line.get_xdata()
            ys = line.get_ydata()
            index = np.abs(xs - x).argmin()
            xlines.append(xs[index])
            ylines.append(ys[index])

        index = np.abs(np.array(ylines) - y).argmin()
        x = xlines[index]
        y = ylines[index]

        crossh, crossv = self._snap_cross[event.inaxes]
        crossh.set_ydata([y, y])
        crossv.set_xdata([x, x])

        s = event.inaxes.format_coord(x, y)
        self.set_message('%s, %s' % (self.mode, s))

        self.draw()

class NavigationToolbarSnapMixinQT(NavigationToolbarSnapMixin):

    def _init_toolbar(self):
        a = QAction(getIcon('snap'), 'Snap', self)
        a.setCheckable(True)
        a.setToolTip('Snap to data')
        a.triggered.connect(self.snap)
        self._actions['snap'] = a
        self.insertAction(self._actions['pan'], a)

    def snap(self, *args):
        super().snap(*args)
        self._update_buttons_checked()

    def _update_buttons_checked(self):
        super()._update_buttons_checked()
        self._actions['snap'].setChecked(self._active == 'SNAP')

class ColorbarWidget(QWidget):

    POSITION_INSIDE = 'inside'
    POSITION_OUTSIDE = 'outside'

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)

        # Variables
        cmaps = sorted(key for key in cm.cmap_d.keys() \
                       if not key.endswith('_r'))

        # Widgets
        self._chk_visible = QCheckBox('Visible')

        self._cb_colormap = QComboBox()
        self._cb_colormap.addItems(cmaps)

        self._chk_reverse = QCheckBox('Reverse')

        self._rb_position = QButtonGroup()
        self._rb_position.addButton(QRadioButton('inside'), 0)
        self._rb_position.addButton(QRadioButton('outside'), 1)

        # Layouts
        layout = QFormLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        layout.addRow(self._chk_visible)
        layout.addRow('Color map', horizontal(self._cb_colormap, self._chk_reverse))

        sublayout = QVBoxLayout()
        for button in self._rb_position.buttons():
            sublayout.addWidget(button)
        layout.addRow('Position', sublayout)

        self.setLayout(layout)

        # Defaults
        self._rb_position.button(0).setChecked(True)

    def colormap(self):
        colormap = self._cb_colormap.currentText()
        if self._chk_reverse.isChecked():
            colormap += '_r'
        return colormap

    def setColormap(self, colormap):
        if isinstance(colormap, colors.Colormap):
            colormap = colormap.name

        if colormap.endswith('_r'):
            colormap = colormap[:-2]
            self._chk_reverse.setChecked(True)
        else:
            self._chk_reverse.setChecked(False)

        index = self._cb_colormap.findText(colormap)
        if index is None:
            raise ValueError('Unknown colormap: %s' % colormap)
        self._cb_colormap.setCurrentIndex(index)

    def colorbarVisible(self):
        return self._chk_visible.isChecked()

    def setColorbarVisible(self, visible):
        self._chk_visible.setChecked(visible)

    def position(self):
        if self._rb_position.button(0).isChecked():
            return self.POSITION_INSIDE
        else:
            return self.POSITION_OUTSIDE

    def setPosition(self, position):
        if position == self.POSITION_INSIDE:
            self._rb_position.button(0).setChecked(True)
        elif position == self.POSITION_OUTSIDE:
            self._rb_position.button(1).setChecked(True)
        else:
            raise ValueError('Unknown position: %s' % position)

class _ColorbarDialog(QDialog):

    def __init__(self, parent=None):
        QDialog.__init__(self, parent)
        self.setWindowTitle('Color bar')

        # Widgets
        self._wdg_colorbar = ColorbarWidget()

        # FIXME: Implement colorbar position
        for button in self._wdg_colorbar._rb_position.buttons():
            button.setEnabled(False)

        buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)

        # Layouts
        layout = QVBoxLayout()

        layout.addWidget(self._wdg_colorbar)
        layout.addWidget(buttons)

        self.setLayout(layout)

        # Signals
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)

    def colormap(self):
        return self._wdg_colorbar.colormap()
        colormap = self._cb_colormap.currentText()
        if self._chk_reverse.isChecked():
            colormap += '_r'
        return colormap

    def setColormap(self, colormap):
        self._wdg_colorbar.setColormap(colormap)

    def colorbarVisible(self):
        return self._wdg_colorbar.colorbarVisible()

    def setColorbarVisible(self, visible):
        self._wdg_colorbar.setColorbarVisible(visible)

    def position(self):
        return self._wdg_colorbar.position()

    def setPosition(self, position):
        self._wdg_colorbar.setPosition(position)

class NavigationToolbarColorbarMixinQT(object):

    colorbar_changed = Signal(bool, str)

    def __init__(self):
        self._colorbar = None

    def _init_toolbar(self):
        a = QAction(getIcon('color-wheel'), 'Color bar', self)
        a.setToolTip('Change color bar')
        a.setEnabled(False)
        a.triggered.connect(self.colorbar)
        self._actions['colorbar'] = a
        self.insertAction(self._actions['configure_subplots'], a)

    def set_colorbar(self, colorbar):
        self._colorbar = colorbar
        self._actions['colorbar'].setEnabled(colorbar is not None)

    def get_colorbar(self):
        return self._colorbar

    def colorbar(self):
        cb = self.get_colorbar()

        dialog = _ColorbarDialog(self)
        dialog.setColorbarVisible(cb.ax.get_visible())
        dialog.setColormap(cb.mappable.get_cmap())
        if not dialog.exec_():
            return

        visible = dialog.colorbarVisible()
        cb.ax.set_visible(visible)

        cmap = dialog.colormap()
        cb.mappable.set_cmap(cmap)

        self.canvas.draw()

        cb = self.get_colorbar()
        cmap = cb.mappable.get_cmap()
        visible = cb.ax.get_visible()

        self.colorbar_changed.emit(visible, cmap.name)

class ScalebarWidget(QWidget):

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)

        # Widgets
        self._chk_visible = QCheckBox('Visible')

        self._cb_location = QComboBox()
        for text, index in sorted(ScaleBar._LOCATIONS.items(), key=itemgetter(1)):
            self._cb_location.addItem(text, index)

        self._sld_length_fraction = QSlider(Qt.Orientation.Horizontal)
        self._sld_length_fraction.setRange(0, 100)
        self._sld_length_fraction.setSingleStep(1)
        self._txt_length_fraction = QLabel()

        self._sld_height_fraction = QSlider(Qt.Orientation.Horizontal)
        self._sld_height_fraction.setRange(0, 100)
        self._sld_height_fraction.setSingleStep(1)
        self._txt_height_fraction = QLabel()

        self._cb_label_top = QComboBox()
        self._cb_label_top.addItem('Top', True)
        self._cb_label_top.addItem('Bottom', False)

        self._btn_color = ColorDialogButton()

        self._btn_box_color = ColorDialogButton()

        self._sld_box_alpha = QSlider(Qt.Orientation.Horizontal)
        self._sld_box_alpha.setRange(0, 100)
        self._sld_box_alpha.setSingleStep(1)
        self._txt_box_alpha = QLabel()

        # Layouts
        layout = QFormLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        layout.addRow(self._chk_visible)
        layout.addRow('Location', self._cb_location)
        layout.addRow("Length of scale bar\n(fraction of width)",
                      horizontal(self._sld_length_fraction,
                                 self._txt_length_fraction))
        layout.addRow("Height of scale bar\n(fraction of height)",
                      horizontal(self._sld_height_fraction,
                                 self._txt_height_fraction))
        layout.addRow('Label location', self._cb_label_top)
        layout.addRow("Text and scale color", self._btn_color)
        layout.addRow("Background color", self._btn_box_color)
        layout.addRow("Background alpha", horizontal(self._sld_box_alpha,
                                                     self._txt_box_alpha))

        self.setLayout(layout)

        # Signals
        self._sld_length_fraction.valueChanged.connect(self._on_length_fraction_changed)
        self._sld_height_fraction.valueChanged.connect(self._on_height_fraction_changed)
        self._sld_box_alpha.valueChanged.connect(self._on_box_alpha_changed)

        # Defaults
        self._sld_length_fraction.setValue(0)
        self._sld_height_fraction.setValue(0)
        self._sld_box_alpha.setValue(0)

    def _on_length_fraction_changed(self, value):
        self._txt_length_fraction.setText('%i%%' % value)

    def _on_height_fraction_changed(self, value):
        self._txt_height_fraction.setText('%i%%' % value)

    def _on_box_alpha_changed(self, value):
        self._txt_box_alpha.setText('%i%%' % value)

    def scalebarVisible(self):
        return self._chk_visible.isChecked()

    def setScalebarVisible(self, visible):
        self._chk_visible.setChecked(visible)

    def location(self):
        index = self._cb_location.currentIndex()
        return self._cb_location.itemData(index)

    def setLocation(self, location):
        index = self._cb_location.findData(location)
        self._cb_location.setCurrentIndex(index)

    def lengthFraction(self):
        return self._sld_length_fraction.value() / 100.0

    def setLengthFraction(self, fraction):
        self._sld_length_fraction.setValue(int(fraction * 100))

    def heightFraction(self):
        return self._sld_height_fraction.value() / 100.0

    def setHeightFraction(self, fraction):
        self._sld_height_fraction.setValue(int(fraction * 100))

    def labelTop(self):
        index = self._cb_label_top.currentIndex()
        return self._cb_label_top.itemData(index)

    def setLabelTop(self, label_top):
        index = self._cb_label_top.findData(label_top)
        self._cb_label_top.setCurrentIndex(index)

    def color(self):
        return self._btn_color.color()

    def setColor(self, color):
        self._btn_color.setColor(color)

    def boxColor(self):
        return self._btn_box_color.color()

    def setBoxColor(self, color):
        self._btn_box_color.setColor(color)

    def boxAlpha(self):
        return self._sld_box_alpha.value() / 100.0

    def setBoxAlpha(self, alpha):
        self._sld_box_alpha.setValue(int(alpha * 100))

class _ScalebarDialog(QDialog):

    def __init__(self, parent=None):
        QDialog.__init__(self, parent)
        self.setWindowTitle('Scale bar')

        # Widgets
        self._wdg_scalebar = ScalebarWidget()

        buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)

        # Layouts
        layout = QVBoxLayout()

        layout.addWidget(self._wdg_scalebar)
        layout.addWidget(buttons)

        self.setLayout(layout)

        # Signals
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)

    def scalebarVisible(self):
        return self._wdg_scalebar.scalebarVisible()

    def setScalebarVisible(self, visible):
        self._wdg_scalebar.setScalebarVisible(visible)

    def location(self):
        return self._wdg_scalebar.location()

    def setLocation(self, location):
        self._wdg_scalebar.setLocation(location)

    def lengthFraction(self):
        return self._wdg_scalebar.lengthFraction()

    def setLengthFraction(self, fraction):
        self._wdg_scalebar.setLengthFraction(fraction)

    def heightFraction(self):
        return self._wdg_scalebar.heightFraction()

    def setHeightFraction(self, fraction):
        self._wdg_scalebar.setHeightFraction(fraction)

    def labelTop(self):
        return self._wdg_scalebar.labelTop()

    def setLabelTop(self, label_top):
        self._wdg_scalebar.setLabelTop(label_top)

    def color(self):
        return self._wdg_scalebar.color()

    def setColor(self, color):
        self._wdg_scalebar.setColor(color)

    def boxColor(self):
        return self._wdg_scalebar.boxColor()

    def setBoxColor(self, color):
        self._wdg_scalebar.setBoxColor(color)

    def boxAlpha(self):
        return self._wdg_scalebar.boxAlpha()

    def setBoxAlpha(self, fraction):
        self._wdg_scalebar.setBoxAlpha(fraction)

class NavigationToolbarScalebarMixinQT(object):

    scalebar_changed = Signal(bool, int, float, float, bool, tuple, tuple, float)

    def __init__(self):
        self._scalebar = None

    def _init_toolbar(self):
        a = QAction(getIcon('ruler'), 'Scale bar', self)
        a.setToolTip('Change scale bar')
        a.setEnabled(False)
        a.triggered.connect(self.scalebar)
        self._actions['scalebar'] = a
        self.insertAction(self._actions['configure_subplots'], a)

    def set_scalebar(self, scalebar):
        self._scalebar = scalebar
        self._actions['scalebar'].setEnabled(scalebar is not None)

    def get_scalebar(self):
        return self._scalebar

    def scalebar(self):
        sb = self.get_scalebar()

        dialog = _ScalebarDialog(self)
        dialog.setScalebarVisible(sb.get_visible())
        dialog.setLocation(sb.location)
        dialog.setLengthFraction(sb.length_fraction)
        dialog.setHeightFraction(sb.height_fraction)
        dialog.setLabelTop(sb.label_top)
        dialog.setColor(convert_color_mpl_to_qt(sb.color))
        dialog.setBoxColor(convert_color_mpl_to_qt(sb.box_color))
        dialog.setBoxAlpha(sb.box_alpha)
        if not dialog.exec_():
            return

        visible = dialog.scalebarVisible()
        location = dialog.location()
        length_fraction = dialog.lengthFraction()
        height_fraction = dialog.heightFraction()
        label_top = dialog.labelTop()
        color = convert_color_qt_to_mpl(dialog.color())
        box_color = convert_color_qt_to_mpl(dialog.boxColor())
        box_alpha = dialog.boxAlpha()

        sb.set_visible(visible)
        sb.location = location
        sb.length_fraction = length_fraction
        sb.height_fraction = height_fraction
        sb.label_top = label_top
        sb.color = color
        sb.box_color = box_color
        sb.box_alpha = box_alpha

        self.canvas.draw()

        self.scalebar_changed.emit(visible, location,
                                   length_fraction, height_fraction,
                                   label_top, color, box_color, box_alpha)




"""
Image raster widgets
"""

# Standard library modules.

# Third party modules.
from PySide.QtGui import \
    QSlider, QFormLayout, QHBoxLayout, QRadioButton, QSplitter, QSizePolicy
from PySide.QtCore import Qt, QAbstractTableModel, Signal

import matplotlib
from matplotlib.figure import Figure

import numpy as np

# Local modules.
from pyhmsa.gui.spec.datum.datum import \
    _DatumWidget, _DatumTableWidget, _DatumFigureWidget
from pyhmsa.gui.spec.datum.analysis import \
    Analysis1DTableWidget, Analysis1DGraphWidget
from pyhmsa.gui.util.mpl.modest_image import imshow as modest_imshow
from pyhmsa.gui.util.mpl.toolbar import \
    (NavigationToolbarQT, NavigationToolbarColorbarMixinQT,
     NavigationToolbarScalebarMixinQT)
from pyhmsa.gui.util.mpl.scalebar import ScaleBar

from pyhmsa.spec.datum.analysis import Analysis1D
from pyhmsa.spec.datum.imageraster import ImageRaster2D, ImageRaster2DSpectral
from pyhmsa.spec.condition.acquisition import AcquisitionRasterXY
from pyhmsa.type.numerical import convert_unit

# Globals and constants variables.

class _ImageRaster2DNavigationToolbarQT(NavigationToolbarColorbarMixinQT,
                                        NavigationToolbarScalebarMixinQT,
                                        NavigationToolbarQT):

    def __init__(self, canvas, parent, coordinates=True):
        NavigationToolbarQT.__init__(self, canvas, parent, coordinates)
        NavigationToolbarColorbarMixinQT.__init__(self)
        NavigationToolbarScalebarMixinQT.__init__(self)

    def _init_toolbar(self):
        NavigationToolbarQT._init_toolbar(self)
        NavigationToolbarColorbarMixinQT._init_toolbar(self)
        NavigationToolbarScalebarMixinQT._init_toolbar(self)

class ImageRaster2DTableWidget(_DatumTableWidget):

    class _TableModel(QAbstractTableModel):

        def __init__(self, datum):
            QAbstractTableModel.__init__(self)
            self._datum = datum

        def rowCount(self, parent=None):
            return self._datum.y

        def columnCount(self, parent=None):
            return self._datum.x

        def data(self, index, role):
            if not index.isValid() or \
                    not (0 <= index.row() < self._datum.y) or \
                    not (0 <= index.column() < self._datum.x):
                return None
            if role != Qt.DisplayRole:
                return None

            row = index.row()
            column = index.column()
            return str(self._datum[column, row])

        def headerData(self, section , orientation, role):
            if role != Qt.DisplayRole:
                return None
            if orientation == Qt.Horizontal:
                return str(section + 1)
            elif orientation == Qt.Vertical:
                return str(section + 1)

    def __init__(self, controller, datum=None, parent=None):
        _DatumTableWidget.__init__(self, ImageRaster2D, controller,
                                   datum, parent)

    def _create_model(self, datum):
        return self._TableModel(datum)

class ImageRaster2DGraphWidget(_DatumFigureWidget):

    def __init__(self, controller, datum=None, parent=None):
        _DatumFigureWidget.__init__(self, ImageRaster2D, controller,
                                    datum, parent)

    def _create_toolbar(self, canvas):
        self._toolbar = _ImageRaster2DNavigationToolbarQT(canvas, self.parent())
        self._toolbar.colorbar_changed.connect(self._on_colorbar_changed)
        self._toolbar.scalebar_changed.connect(self._on_scalebar_changed)
        return self._toolbar

    def _create_figure(self):
        fig = Figure((5.0, 4.0), dpi=100)
        self._ax = None
        self._artist = None
        return fig

    def _create_colorbar(self, fig):
        settings = self.controller.settings

        colorbar = fig.colorbar(self._artist, ax=self._ax, shrink=0.8)
        colorbar.ax.set_visible(settings.valueBool('colorbar/visible', True))

        self._toolbar.set_colorbar(colorbar)

        return colorbar

    def _create_scalebar(self, datum):
        acqs = datum.conditions.findvalues(AcquisitionRasterXY)
        if not acqs:
            return None
        acq = next(iter(acqs))
        dx_m = convert_unit('m', acq.step_size_x)

        settings = self.controller.settings
        visible = settings.valueBool('scalebar/visible', True)
        length_fraction = float(settings.value('scalebar/lengthFraction', 0.2))
        height_fraction = float(settings.value('scalebar/heightFraction', 0.01))
        location = int(settings.value('scalebar/location', 3))
        color = settings.valueColor('scalebar/color', 'w')
        box_color = settings.valueColor('scalebar/boxColor', 'k')
        box_alpha = float(settings.value('scalebar/boxAlpha', 0.5))
        label_top = settings.valueBool('scalebar/labelTop', False)

        scalebar = ScaleBar(dx_m,
                            length_fraction=length_fraction,
                            height_fraction=height_fraction,
                            location=location,
                            color=color,
                            box_color=box_color,
                            box_alpha=box_alpha,
                            label_top=label_top)
        scalebar.set_visible(visible)

        self._toolbar.set_scalebar(scalebar)

        return scalebar

    def _draw_figure(self, fig, datum):
        if datum is None:
            self._ax = None
            self._artist = None
            return

        if self._artist is None: # First draw
            self._ax = fig.add_subplot("111")
            fig.subplots_adjust(0.0, 0.0, 1.0, 1.0)

            self._ax.xaxis.set_visible(False)
            self._ax.yaxis.set_visible(False)

            settings = self.controller.settings
            cmap = settings.value('colorbar/colormap', 'jet') or 'jet'
            self._artist = modest_imshow(self._ax, datum.T, cmap=cmap)

            scalebar = self._create_scalebar(datum)
            if scalebar is not None:
                self._ax.add_artist(scalebar)

            self._create_colorbar(fig)

        else:
            self._artist.set_data(datum.T)
            self._artist.autoscale()

    def _on_colorbar_changed(self, visible, colormap):
        settings = self.controller.settings
        settings.setValue('colorbar/visible', visible)
        settings.setValue('colorbar/colormap', colormap)

    def _on_scalebar_changed(self, visible, location,
                             length_fraction, height_fraction,
                             label_top, color, box_color, box_alpha):
        settings = self.controller.settings
        settings.setValue('scalebar/visible', visible)
        settings.setValue('scalebar/location', location)
        settings.setValue('scalebar/lengthFraction', length_fraction)
        settings.setValue('scalebar/heightFraction', height_fraction)
        settings.setValueColor('scalebar/color', color)
        settings.setValueColor('scalebar/boxColor', box_color)
        settings.setValue('scalebar/boxAlpha', box_alpha)
        settings.setValue('scalebar/labelTop', label_top)

class _ImageRaster2DSpectralWidget(_DatumWidget):

    MODE_SUM = 'sum'
    MODE_MAX = 'max'
    MODE_SINGLE = 'single'
    MODE_RANGE = 'range'

    valueSelected = Signal(int, int)

    def __init__(self, controller, datum=None, parent=None):
        self._datum = None
        _DatumWidget.__init__(self, ImageRaster2DSpectral, controller,
                              datum, parent)

    def _init_ui(self):
        # Widgets
        self._rdb_sum = QRadioButton("Sum")
        self._rdb_sum.setChecked(True)
        self._rdb_max = QRadioButton("Maximum")
        self._rdb_single = QRadioButton("Single")
        self._rdb_range = QRadioButton("Range")

        self._sld_start = QSlider(Qt.Orientation.Horizontal)
        self._sld_start.setTickPosition(QSlider.TicksBelow)
        self._sld_start.setEnabled(False)

        self._sld_end = QSlider(Qt.Orientation.Horizontal)
        self._sld_end.setTickPosition(QSlider.TicksBelow)
        self._sld_end.setEnabled(False)

        self._wdg_imageraster2d = self._create_imageraster2d_widget()
        self._wdg_analysis = self._create_analysis1d_widget()

        # Layouts
        layout = _DatumWidget._init_ui(self)

        sublayout = QHBoxLayout()
        sublayout.addWidget(self._rdb_sum)
        sublayout.addWidget(self._rdb_max)
        sublayout.addWidget(self._rdb_single)
        sublayout.addWidget(self._rdb_range)
        layout.addLayout(sublayout)

        sublayout = QFormLayout()
        sublayout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow) # Fix for Mac OS
        sublayout.addRow('Channels (Start)', self._sld_start)
        sublayout.addRow('Channels (End)', self._sld_end)
        layout.addLayout(sublayout)

        splitter = QSplitter()
        splitter.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        splitter.addWidget(self._wdg_imageraster2d)
        splitter.addWidget(self._wdg_analysis)
        layout.addWidget(splitter)

        # Signals
        self._rdb_sum.toggled.connect(self._on_mode_sum)
        self._rdb_max.toggled.connect(self._on_mode_max)
        self._rdb_single.toggled.connect(self._on_mode_single)
        self._rdb_range.toggled.connect(self._on_mode_range)

        self._sld_start.valueChanged.connect(self._on_slide_start)
        self._sld_end.valueChanged.connect(self._on_slide_end)

        self.valueSelected.connect(self._on_value_selected)

        # Defaults
        self.setMode(self.MODE_SUM)

        return layout

    def _create_analysis1d_widget(self):
        raise NotImplementedError

    def _create_imageraster2d_widget(self):
        raise NotImplementedError

    def _on_mode_sum(self, checked):
        if checked:
            self.setMode(self.MODE_SUM)

    def _on_mode_max(self, checked):
        if checked:
            self.setMode(self.MODE_MAX)

    def _on_mode_single(self, checked):
        if checked:
            self.setMode(self.MODE_SINGLE)

    def _on_mode_range(self, checked):
        if checked:
            self.setMode(self.MODE_RANGE)

    def _update_data(self, mode=None):
        if mode is None:
            mode = self.mode()

        if mode == self.MODE_SUM:
            self._update_mode_sum()
        elif mode == self.MODE_MAX:
            self._update_mode_max()
        elif mode == self.MODE_SINGLE:
            self._update_mode_single()
        elif mode == self.MODE_RANGE:
            self._update_mode_range()

    def _update_mode_sum(self):
        if self._datum is None:
            return
        subdatum = np.sum(self._datum, 2)
        self._wdg_imageraster2d.setDatum(subdatum)

    def _update_mode_max(self):
        if self._datum is None:
            return
        subdatum = np.amax(self._datum, 2)
        self._wdg_imageraster2d.setDatum(subdatum)

    def _update_mode_single(self):
        if self._datum is None:
            return
        channel = self._sld_start.value()
        subdatum = self._datum[:, :, channel]
        self._wdg_imageraster2d.setDatum(subdatum)

    def _update_mode_range(self):
        if self._datum is None:
            return
        start = self._sld_start.value()
        end = self._sld_end.value()
        start2 = min(start, end)
        end2 = max(start, end)
        subdatum = np.sum(self._datum[:, :, start2:end2 + 1], 2)
        self._wdg_imageraster2d.setDatum(subdatum)

    def _on_slide_start(self, channel):
        if self._rdb_single.isChecked():
            self._update_mode_single()
        elif self._rdb_range.isChecked():
            self._update_mode_range()

    def _on_slide_end(self, channel):
        self._update_mode_range()

    def _on_value_selected(self, x, y):
        if self._datum is None:
            return
        subdatum = self._datum[x, y]
        self._wdg_analysis.setDatum(subdatum.view(Analysis1D))
        self._update_data()

    def setDatum(self, datum):
        _DatumWidget.setDatum(self, datum)
        self._datum = datum

        self._sld_start.setMaximum(datum.channels - 1)
        self._sld_end.setMaximum(datum.channels - 1)

        self._update_data()

    def setMode(self, mode):
        rsum = rmax = rsingle = rrange = False
        sstart = send = False
        if mode == self.MODE_SUM:
            rsum = True
        elif mode == self.MODE_MAX:
            rmax = True
        elif mode == self.MODE_SINGLE:
            rsingle = True
            sstart = True
        elif mode == self.MODE_RANGE:
            rrange = True
            sstart = send = True
        else:
            raise ValueError('Unknown mode')

        self._rdb_sum.setChecked(rsum)
        self._rdb_max.setChecked(rmax)
        self._rdb_single.setChecked(rsingle)
        self._rdb_range.setChecked(rrange)
        self._sld_start.setEnabled(sstart)
        self._sld_end.setEnabled(send)

        self._update_data(mode)

    def mode(self):
        if self._rdb_sum.isChecked():
            return self.MODE_SUM
        elif self._rdb_max.isChecked():
            return self.MODE_MAX
        elif self._rdb_single.isChecked():
            return self.MODE_SINGLE
        elif self._rdb_range.isChecked():
            return self.MODE_RANGE
        else:
            raise ValueError('Unknown mode')

class ImageRaster2DSpectralTableWidget(_ImageRaster2DSpectralWidget):

    def _create_imageraster2d_widget(self):
        widget = ImageRaster2DTableWidget(self.controller)
        widget._table.clicked.connect(self._on_table_clicked)
        return widget

    def _create_analysis1d_widget(self):
        return Analysis1DTableWidget(self.controller)

    def _on_table_clicked(self, index):
        self.valueSelected.emit(index.column(), index.row())

class _Analysis1DGraphWidget2(Analysis1DGraphWidget):

    def _create_figure(self):
        fig = Analysis1DGraphWidget._create_figure(self)
        self._ax_single = None
        self._ax_range = None
        return fig

    def _draw_figure(self, fig, datum):
        Analysis1DGraphWidget._draw_figure(self, fig, datum)

        if datum is None:
            self._ax_single = None
            self._ax_range = None
            return

        color = matplotlib.rcParams['axes.color_cycle'][1]

        self._ax_single = self._ax.axvline(0, lw=3, color=color, zorder=3)
        self._ax_single.set_visible(False)

        self._ax_range = self._ax.axvspan(0, 0, alpha=0.5, facecolor=color, zorder=3)
        self._ax_range.set_visible(False)

    def setMode(self, mode):
        if self._ax_single is None or self._ax_range is None:
            return

        vsingle = vrange = False
        if mode == _ImageRaster2DSpectralWidget.MODE_SINGLE:
            vsingle = True
        elif mode == _ImageRaster2DSpectralWidget.MODE_RANGE:
            vrange = True

        self._ax_single.set_visible(vsingle)
        self._ax_range.set_visible(vrange)

        self._canvas.draw()

    def setSinglePosition(self, channel):
        if self._artist is None or self._ax_single is None:
            return

        self._ax_single.set_visible(True)

        xs = self._artist.get_xdata()
        ys = self._artist.get_ydata()

        x = xs[channel]
        ymin = min(ys); ymax = max(ys)
        self._ax_single.set_xdata([x, x])
        self._ax_single.set_ydata([ymin, ymax])

        self._canvas.draw()

    def setRange(self, start, end):
        if self._artist is None or self._ax_range is None:
            return

        self._ax_range.set_visible(True)

        xs = self._artist.get_xdata()
        ys = self._artist.get_ydata()

        xmin = xs[start]; xmax = xs[end]
        ymin = min(ys); ymax = max(ys)
        self._ax_range.set_xy([[xmin, ymin],
                               [xmin, ymax],
                               [xmax, ymax],
                               [xmax, ymin]])

        self._canvas.draw()

class ImageRaster2DSpectralGraphWidget(_ImageRaster2DSpectralWidget):

    def _create_imageraster2d_widget(self):
        widget = ImageRaster2DGraphWidget(self.controller)
        widget._canvas.mpl_connect("button_release_event", self._onFigureClicked)
        return widget

    def _create_analysis1d_widget(self):
        return _Analysis1DGraphWidget2(self.controller)

    def _onFigureClicked(self, event):
        if not event.inaxes:
            return
        self.valueSelected.emit(event.xdata, event.ydata)

    def _update_mode_single(self):
        _ImageRaster2DSpectralWidget._update_mode_single(self)

        channel = self._sld_start.value()
        self._wdg_analysis.setSinglePosition(channel)

    def _update_mode_range(self):
        _ImageRaster2DSpectralWidget._update_mode_range(self)

        start = self._sld_start.value()
        end = self._sld_end.value()
        self._wdg_analysis.setRange(start, end)

    def setMode(self, mode):
        _ImageRaster2DSpectralWidget.setMode(self, mode)
        self._wdg_analysis.setMode(mode)





"""
Utilities for color between matplotlib and PySide
"""

# Standard library modules.

# Third party modules.
from PySide.QtGui import QColor

import matplotlib.colors as colors

# Local modules.

# Globals and constants variables.

def convert_color_qt_to_mpl(color):
    return (color.red() / 255.0, color.green() / 255.0, color.blue() / 255.0)

def convert_color_mpl_to_qt(color):
    r, g, b = colors.colorConverter.to_rgb(color)
    return QColor(int(r * 255), int(g * 255), int(b * 255))


"""
Instrument widget
"""

# Standard library modules.

# Third party modules.

# Local modules.
from pyhmsa.gui.spec.condition.condition import _ConditionWidget
from pyhmsa.gui.util.parameter import TextAttributeLineEdit

from pyhmsa.spec.condition.instrument import Instrument

# Globals and constants variables.

class InstrumentWidget(_ConditionWidget):

    def __init__(self, parent=None):
        _ConditionWidget.__init__(self, Instrument, parent)

    def _init_ui(self):
        # Controls
        self._txt_manufacturer = TextAttributeLineEdit(self.CLASS.manufacturer)
        self._txt_model = TextAttributeLineEdit(self.CLASS.model)
        self._txt_serial_number = TextAttributeLineEdit(self.CLASS.serial_number)

        # Layouts
        layout = _ConditionWidget._init_ui(self)
        layout.addRow("<i>Manufacturer</i>", self._txt_manufacturer)
        layout.addRow("<i>Model</i>", self._txt_model)
        layout.addRow("Serial number", self._txt_serial_number)

        # Signals
        self._txt_manufacturer.textEdited.connect(self.edited)
        self._txt_model.textEdited.connect(self.edited)
        self._txt_serial_number.textEdited.connect(self.edited)

        return layout

    def _create_parameter(self):
        return self.CLASS('manufacturer', 'model')

    def parameter(self, parameter=None):
        parameter = _ConditionWidget.parameter(self, parameter)
        parameter.manufacturer = self._txt_manufacturer.text()
        parameter.model = self._txt_model.text()
        parameter.serial_number = self._txt_serial_number.text()
        return parameter

    def setParameter(self, condition):
        _ConditionWidget.setParameter(self, condition)
        self._txt_manufacturer.setText(condition.manufacturer)
        self._txt_model.setText(condition.model)
        self._txt_serial_number.setText(condition.serial_number)

    def setReadOnly(self, state):
        _ConditionWidget.setReadOnly(self, state)
        self._txt_manufacturer.setReadOnly(state)
        self._txt_model.setReadOnly(state)
        self._txt_serial_number.setReadOnly(state)

    def isReadOnly(self):
        return _ConditionWidget.isReadOnly(self) and \
            self._txt_manufacturer.isReadOnly() and \
            self._txt_model.isReadOnly() and \
            self._txt_serial_number.isReadOnly()

    def hasAcceptableInput(self):
        return _ConditionWidget.hasAcceptableInput(self) and \
            self._txt_manufacturer.hasAcceptableInput() and \
            self._txt_model.hasAcceptableInput() and \
            self._txt_serial_number.hasAcceptableInput()





"""
Special message box
"""

# Standard library modules.
import sys
import traceback
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

# Third party modules.
from PySide.QtGui import QMessageBox

# Local modules.

# Globals and constants variables.

def exception(parent, ex, buttons=QMessageBox.Ok,
              defaultButton=QMessageBox.NoButton):
    title = type(ex).__name__
    message = str(ex)
    tb = StringIO()
    if hasattr(ex, '__traceback__'):
        exc_traceback = ex.__traceback__
    else:
        exc_traceback = sys.exc_info()[2]
    traceback.print_tb(exc_traceback, file=tb)

    msgbox = QMessageBox(QMessageBox.Icon.Critical, title, message, buttons, parent)
    msgbox.setDefaultButton(defaultButton)
    msgbox.setDetailedText(tb.getvalue())
    msgbox.exec_()

def exceptions(parent, exs, buttons=QMessageBox.Ok,
              defaultButton=QMessageBox.NoButton):
    title = 'Exception(s)'
    message = '\n'.join(map(str, exs))

    tracebacks = []
    for ex in exs:
        tb = StringIO()
        if not hasattr(ex, '__traceback__'):
            continue
        exc_traceback = ex.__traceback__
        traceback.print_tb(exc_traceback, file=tb)
        tracebacks.append(tb.getvalue())

    msgbox = QMessageBox(QMessageBox.Icon.Critical, title, message, buttons, parent)
    msgbox.setDefaultButton(defaultButton)
    msgbox.setDetailedText('\n'.join(tracebacks))
    msgbox.exec_()



"""
Region of interest widget
"""

# Standard library modules.

# Third party modules.
from PySide.QtGui import QLineEdit, QIntValidator

# Local modules.
from pyhmsa.gui.spec.condition.condition import _ConditionWidget

from pyhmsa.spec.condition.region import RegionOfInterest

# Globals and constants variables.

class RegionOfInterestWidget(_ConditionWidget):

    def __init__(self, parent=None):
        _ConditionWidget.__init__(self, RegionOfInterest, parent)

    def _init_ui(self):
        # Widgets
        validator = QIntValidator()
        validator.setBottom(0)

        self._txt_start_channel = QLineEdit()
        self._txt_start_channel.setValidator(validator)
        self._txt_start_channel.setAccessibleName('Start channel')

        self._txt_end_channel = QLineEdit()
        self._txt_end_channel.setValidator(validator)
        self._txt_end_channel.setAccessibleName("End channel")

        # Layouts
        layout = _ConditionWidget._init_ui(self)
        layout.addRow('<i>Start channel</i>', self._txt_start_channel)
        layout.addRow('<i>End channel</i>', self._txt_end_channel)

        # Signals
        self._txt_start_channel.textEdited.connect(self.edited)
        self._txt_end_channel.textEdited.connect(self.edited)

        return layout

    def _create_parameter(self):
        return self.CLASS(0, 1)

    def parameter(self, parameter=None):
        parameter = _ConditionWidget.parameter(self, parameter)
        parameter.channels = (int(self._txt_start_channel.text()),
                              int(self._txt_end_channel.text()))
        return parameter

    def setParameter(self, condition):
        _ConditionWidget.setParameter(self, condition)
        self._txt_start_channel.setText(str(condition.start_channel))
        self._txt_end_channel.setText(str(condition.end_channel))

    def setReadOnly(self, state):
        _ConditionWidget.setReadOnly(self, state)
        self._txt_start_channel.setReadOnly(state)
        self._txt_end_channel.setReadOnly(state)

    def isReadOnly(self):
        return _ConditionWidget.isReadOnly(self) and \
            self._txt_start_channel.isReadOnly() and \
            self._txt_end_channel.isReadOnly()

    def hasAcceptableInput(self):
        return _ConditionWidget.hasAcceptableInput(self) and \
            self._txt_start_channel.hasAcceptableInput() and \
            self._txt_end_channel.hasAcceptableInput()



"""
Periodic table of element dialog
"""

# Standard library modules.
import math

# Third party modules.
from PySide.QtGui import \
    (QWidget, QDialog, QPushButton, QPainter, QGridLayout, QColor, QFont,
     QLabel, QSizePolicy, QButtonGroup, QDialogButtonBox, QVBoxLayout,
     QMessageBox)
from PySide.QtCore import Qt, QSize, Signal

import six

# Local modules.
from pyhmsa.util.element_properties import get_symbol, get_atomic_number

# Globals and constants variables.

class ElementPushButton(QPushButton):

    def __init__(self, atomic_number, parent=None):
        QPushButton.__init__(self, parent)

        self._atomic_number = atomic_number
        self._symbol = get_symbol(atomic_number)

        self.setText(self._symbol)

        font = self.font()
        font.setWeight(QFont.Bold)
        self.setFont(font)

        self.setSizePolicy(QSizePolicy.MinimumExpanding,
                           QSizePolicy.MinimumExpanding)

        self.setDefault(False)
        self.setAutoDefault(False)

    def paintEvent(self, event):
        QPushButton.paintEvent(self, event)

        qp = QPainter()
        qp.begin(self)

        font = qp.font()
        font.setPixelSize(self.font().pixelSize() * 0.6)
        font.setWeight(QFont.Normal)
        qp.setFont(font)

        qp.drawText(event.rect().translated(2, 2), str(self._atomic_number))

        qp.end()

    def resizeEvent(self, event):
        font = self.font()
        font.setPixelSize(event.size().height() * 0.4)
        self.setFont(font)
        return QPushButton.resizeEvent(self, event)

    def sizeHint(self, *args, **kwargs):
        return QSize(40, 40)

    @property
    def atomic_number(self):
        return self._atomic_number

    @property
    def symbol(self):
        return self._symbol

_ELEMENT_POSITIONS = \
    {1: (0, 0), 2: (0, 17), 3: (1, 0), 4: (1, 1), 5: (1, 12), 6: (1, 13),
     7: (1, 14), 8: (1, 15), 9: (1, 16), 10: (1, 17), 11: (2, 0), 12: (2, 1),
     13: (2, 12), 14: (2, 13), 15: (2, 14), 16: (2, 15), 17: (2, 16),
     18: (2, 17), 19: (3, 0), 20: (3, 1), 21: (3, 2), 22: (3, 3), 23: (3, 4),
     24: (3, 5), 25: (3, 6), 26: (3, 7), 27: (3, 8), 28: (3, 9), 29: (3, 10),
     30: (3, 11), 31: (3, 12), 32: (3, 13), 33: (3, 14), 34: (3, 15),
     35: (3, 16), 36: (3, 17), 37: (4, 0), 38: (4, 1), 39: (4, 2), 40: (4, 3),
     41: (4, 4), 42: (4, 5), 43: (4, 6), 44: (4, 7), 45: (4, 8), 46: (4, 9),
     47: (4, 10), 48: (4, 11), 49: (4, 12), 50: (5, 12), 51: (4, 14),
     52: (4, 15), 53: (4, 16), 54: (4, 17), 55: (5, 0), 56: (5, 1), 57: (8, 3),
     58: (8, 4), 59: (8, 5), 60: (8, 6), 61: (8, 7), 62: (8, 8), 63: (8, 9),
     64: (8, 10), 65: (8, 11), 66: (8, 12), 67: (8, 13), 68: (8, 14),
     69: (8, 15), 70: (8, 16), 71: (8, 17), 72: (5, 3), 73: (5, 4), 74: (5, 5),
     75: (5, 6), 76: (5, 7), 77: (5, 8), 78: (5, 9), 79: (5, 10), 80: (5, 11),
     81: (4, 13), 82: (5, 13), 83: (5, 14), 84: (5, 15), 85: (5, 16),
     86: (5, 17), 87: (6, 0), 88: (6, 1), 89: (9, 3), 90: (9, 4), 91: (9, 5),
     92: (9, 6), 93: (9, 7), 94: (9, 8), 95: (9, 9), 96: (9, 10), 97: (9, 11),
     98: (9, 12), 99: (9, 13), 100: (9, 14), 101: (9, 15), 102: (9, 16),
     103: (9, 17), 104: (6, 3), 105: (6, 4), 106: (6, 5), 107: (6, 6),
     108: (6, 7), 109: (6, 8), 110: (6, 9), 111: (6, 10), 112: (6, 11),
     113: (6, 12), 114: (6, 13), 115: (6, 14), 116: (6, 15), 117: (6, 16),
     118: (6, 17)}

def _category_color_function(z):
    ALKALI_METALS = QColor(204, 153, 204)
    ALKALI_EARTH_METALS = QColor(0, 255, 255)
    NON_METALS = QColor(0, 204, 102)
    TRANSITION_METALS = QColor(102, 255, 153)
    OTHER_METALS = QColor(153, 51, 204)
    HALOGENS = QColor(255, 204, 0)
    INERT_GAS = QColor(204, 0, 0)
    LANTHANIDES = QColor(102, 0, 255)
    ACTINIDES = QColor(102, 102, 255)
    METALLOIDS = QColor(80, 150, 150)

    if z in [1]:
        return NON_METALS

    elif z in [3, 11, 19, 37, 55, 87]:
        return ALKALI_METALS

    elif z in [4, 12, 20, 38, 56, 88]:
        return ALKALI_EARTH_METALS

    elif z in [21, 22, 23, 24, 25, 26, 27, 28, 29, 30,
               39, 40, 41, 42, 43, 44, 45, 46, 47, 48,
               72, 73, 74, 75, 76, 77, 78, 79, 80,
               104, 105, 106, 107, 108, 109, 110, 111, 112]:
        return TRANSITION_METALS

    elif z in [5, 14, 32, 33, 51, 52, 84]:
        return METALLOIDS

    elif z in [6, 7, 8, 14, 15, 16, 33, 34, 52]:
        return NON_METALS

    elif z in [13, 31, 49, 50, 81, 82, 83, 113, 114, 115, 116]:
        return OTHER_METALS

    elif z in [9, 17, 35, 53, 85, 117]:
        return HALOGENS

    elif z in [2, 10, 18, 36, 54, 86, 118]:
        return INERT_GAS

    elif z in [57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71]:
        return LANTHANIDES

    elif z in [89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103]:
        return ACTINIDES

    raise ValueError('No color definition for z: %s' % z)

def _calculate_brightness(color):
    """
    Returns a value between 0 and 255 corresponding to the brightness of the
    specified color.

    From:http://tech.chitgoks.com/2010/07/27/check-if-color-is-dark-or-light-using-java/
    """
    return int(math.sqrt(color.red() ** 2 * .241 + \
                         color.green() ** 2 * .691 + \
                         color.blue() ** 2 * .068))

class PeriodicTableWidget(QWidget):

    selectionChanged = Signal()

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)

        # Widgets, layouts and signals
        self._group = QButtonGroup()

        layout = QGridLayout()
        layout.setSpacing(0)

        ## Element
        for z, position in _ELEMENT_POSITIONS.items():
            widget = ElementPushButton(z)
            widget.setCheckable(True)
            layout.addWidget(widget, *position)
            self._group.addButton(widget, z)

        ## Labels
        layout.addWidget(QLabel(''), 7, 0) # Dummy
        layout.addWidget(QLabel('*'), 5, 2, Qt.AlignRight)
        layout.addWidget(QLabel('*'), 8, 2, Qt.AlignRight)
        layout.addWidget(QLabel('**'), 6, 2, Qt.AlignRight)
        layout.addWidget(QLabel('**'), 9, 2, Qt.AlignRight)

        for row in [0, 1, 2, 3, 4, 5, 6, 8, 9]:
            layout.setRowStretch(row, 1)

        self.setLayout(layout)

        # Signals
        self._group.buttonClicked.connect(self.selectionChanged)

        # Default
        self.setColorFunction(_category_color_function)

    def setColorFunction(self, func):
        if not callable(func):
            raise ValueError('Not a function')
        self._color_function = func

        # Redraw
        for widget in self._group.buttons():
            z = self._group.id(widget)
            bcolor = func(z)
            fcolor = 'white' if _calculate_brightness(bcolor) < 128 else 'black'
            sheet = 'background-color: %s; color: %s' % (bcolor.name(), fcolor)
            widget.setStyleSheet(sheet)

    def colorFunction(self):
        return self._color_function

    def setMultipleSelection(self, multiple):
        self._group.setExclusive(not multiple)

    def isMultipleSelection(self):
        return not self._group.exclusive()

    def setSelection(self, selection):
        def _uncheckedAll():
            for widget in self._group.buttons():
                widget.setChecked(False)

        if selection is None:
            _uncheckedAll()
            self.selectionChanged.emit()
            return

        if isinstance(selection, (int, six.string_types)):
            selection = [selection]

        if not self.isMultipleSelection() and len(selection) > 1:
            raise ValueError('Multiple selection mode is off. Cannot select more than one element')

        _uncheckedAll()

        for z in selection:
            if isinstance(z, six.string_types):
                z = get_atomic_number(z)
            self._group.button(z).setChecked(True)

        self.selectionChanged.emit()
#
    def selection(self):
        selection = set()
        for widget in self._group.buttons():
            if widget.isChecked():
                selection.add(self._group.id(widget))

        if self.isMultipleSelection():
            return frozenset(selection)
        else:
            if len(selection) > 0:
                return list(selection)[0]
            else:
                return None

    def selectionSymbol(self):
        selection = self.selection()
        if self.isMultipleSelection():
            return frozenset(map(get_symbol, selection))
        else:
            if selection is None:
                return None
            else:
                return get_symbol(selection)

class PeriodicTableDialog(QDialog):

    selectionChanged = Signal()

    def __init__(self, parent=None):
        QDialog.__init__(self, parent)
        self.setWindowTitle('Periodic table')

        # Variables
        self._required_selection = True

        # Widgets
        self._wdg_table = PeriodicTableWidget()

        buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)

        # Layouts
        layout = QVBoxLayout()
        layout.addWidget(self._wdg_table)
        layout.addWidget(buttons)
        self.setLayout(layout)

        # Signals
        self._wdg_table.selectionChanged.connect(self.selectionChanged)
        buttons.accepted.connect(self._onOk)
        buttons.rejected.connect(self._onCancel)

    def _onOk(self):
        if self.isRequiresSelection() and not self._wdg_table.selection():
            QMessageBox.critical(self, self.windowTitle(),
                                 'Please select one element')
            return
        self.accept()

    def _onCancel(self):
        self.setSelection(None)
        self.reject()

    def setColorFunction(self, func):
        self._wdg_table.setColorFunction(func)

    def colorFunction(self):
        return self._wdg_table.colorFunction()

    def setMultipleSelection(self, multiple):
        self._wdg_table.setMultipleSelection(multiple)

    def isMultipleSelection(self):
        return self._wdg_table.isMultipleSelection()

    def setSelection(self, zs):
        self._wdg_table.setSelection(zs)
#
    def selection(self):
        return self._wdg_table.selection()

    def selectionSymbol(self):
        return self._wdg_table.selectionSymbol()

    def setRequiresSelection(self, required):
        self._required_selection = required

    def isRequiresSelection(self):
        return self._required_selection

def run():
    import sys
    from PySide.QtGui import QApplication

    def selection():
        print('hello')

    app = QApplication(sys.argv)
    dialog = PeriodicTableDialog(None)
    dialog.selectionChanged.connect(selection)
    dialog.exec_()
    print(dialog.selection())
    app.exec_()

if __name__ == '__main__':
    run()



"""
Probe widgets
"""

# Standard library modules.

# Third party modules.
from PySide.QtGui import QComboBox

# Local modules.
from pyhmsa.gui.spec.condition.condition import _ConditionWidget
from pyhmsa.gui.util.parameter import NumericalAttributeLineEdit

from pyhmsa.spec.condition.probe import ProbeEM, ProbeTEM

# Globals and constants variables.
from pyhmsa.spec.condition.probe import _GUN_TYPES, _LENS_MODES, LENS_MODE_IMAGE

class ProbeEMWidget(_ConditionWidget):

    def __init__(self, parent=None):
        _ConditionWidget.__init__(self, ProbeEM, parent)

    def _init_ui(self):
        # Widgets
        self._txt_beam_voltage = NumericalAttributeLineEdit(self.CLASS.beam_voltage)
        self._txt_beam_current = NumericalAttributeLineEdit(self.CLASS.beam_current)
        self._cb_gun_type = QComboBox()
        self._cb_gun_type.addItems([None] + list(_GUN_TYPES))
        self._txt_emission_current = NumericalAttributeLineEdit(self.CLASS.emission_current)
        self._txt_filament_current = NumericalAttributeLineEdit(self.CLASS.filament_current)
        self._txt_extractor_bias = NumericalAttributeLineEdit(self.CLASS.extractor_bias)
        self._txt_beam_diameter = NumericalAttributeLineEdit(self.CLASS.beam_diameter)
        self._txt_chamber_pressure = NumericalAttributeLineEdit(self.CLASS.chamber_pressure)
        self._txt_gun_pressure = NumericalAttributeLineEdit(self.CLASS.gun_pressure)
        self._txt_scan_magnification = NumericalAttributeLineEdit(self.CLASS.scan_magnification)
        self._txt_scan_magnification.setFormat('{0:d}')
        self._txt_working_distance = NumericalAttributeLineEdit(self.CLASS.working_distance)

        # Layout
        layout = _ConditionWidget._init_ui(self)
        layout.addRow('<i>Beam voltage</i>', self._txt_beam_voltage)
        layout.addRow('Beam current', self._txt_beam_current)
        layout.addRow('Type of gun', self._cb_gun_type)
        layout.addRow('Emission current', self._txt_emission_current)
        layout.addRow('Filament current', self._txt_filament_current)
        layout.addRow('Extractor bias', self._txt_extractor_bias)
        layout.addRow('Beam diameter', self._txt_beam_diameter)
        layout.addRow('Chamber pressure', self._txt_chamber_pressure)
        layout.addRow('Gun pressure', self._txt_gun_pressure)
        layout.addRow('Scan magnification', self._txt_scan_magnification)
        layout.addRow('Working distance', self._txt_working_distance)

        # Signals
        self._txt_beam_voltage.textEdited.connect(self.edited)
        self._txt_beam_current.textEdited.connect(self.edited)
        self._cb_gun_type.currentIndexChanged.connect(self.edited)
        self._txt_emission_current.textEdited.connect(self.edited)
        self._txt_filament_current.textEdited.connect(self.edited)
        self._txt_extractor_bias.textEdited.connect(self.edited)
        self._txt_beam_diameter.textEdited.connect(self.edited)
        self._txt_chamber_pressure.textEdited.connect(self.edited)
        self._txt_gun_pressure.textEdited.connect(self.edited)
        self._txt_scan_magnification.textEdited.connect(self.edited)
        self._txt_working_distance.textEdited.connect(self.edited)

        return layout

    def _create_parameter(self):
        return self.CLASS(1e3)

    def parameter(self, parameter=None):
        parameter = _ConditionWidget.parameter(self, parameter)
        parameter.beam_voltage = self._txt_beam_voltage.text()
        parameter.beam_current = self._txt_beam_current.text()
        parameter.gun_type = self._cb_gun_type.currentText()
        parameter.emission_current = self._txt_emission_current.text()
        parameter.filament_current = self._txt_filament_current.text()
        parameter.extractor_bias = self._txt_extractor_bias.text()
        parameter.beam_diameter = self._txt_beam_diameter.text()
        parameter.chamber_pressure = self._txt_chamber_pressure.text()
        parameter.gun_pressure = self._txt_gun_pressure.text()
        parameter.scan_magnification = self._txt_scan_magnification.text()
        parameter.working_distance = self._txt_working_distance.text()
        return parameter

    def setParameter(self, condition):
        _ConditionWidget.setParameter(self, condition)
        self._txt_beam_voltage.setText(condition.beam_voltage)
        self._txt_beam_current.setText(condition.beam_current)
        self._cb_gun_type.setCurrentIndex(self._cb_gun_type.findText(condition.gun_type))
        self._txt_emission_current.setText(condition.emission_current)
        self._txt_filament_current.setText(condition.filament_current)
        self._txt_extractor_bias.setText(condition.extractor_bias)
        self._txt_beam_diameter.setText(condition.beam_diameter)
        self._txt_chamber_pressure.setText(condition.chamber_pressure)
        self._txt_gun_pressure.setText(condition.gun_pressure)
        self._txt_scan_magnification.setText(condition.scan_magnification)
        self._txt_working_distance.setText(condition.working_distance)

    def setReadOnly(self, state):
        _ConditionWidget.setReadOnly(self, state)
        self._txt_beam_voltage.setReadOnly(state)
        self._txt_beam_current.setReadOnly(state)
        self._cb_gun_type.setEnabled(not state)
        self._txt_emission_current.setReadOnly(state)
        self._txt_filament_current.setReadOnly(state)
        self._txt_extractor_bias.setReadOnly(state)
        self._txt_beam_diameter.setReadOnly(state)
        self._txt_chamber_pressure.setReadOnly(state)
        self._txt_gun_pressure.setReadOnly(state)
        self._txt_scan_magnification.setReadOnly(state)
        self._txt_working_distance.setReadOnly(state)

    def isReadOnly(self):
        return _ConditionWidget.isReadOnly(self) and \
            self._txt_beam_voltage.isReadOnly() and \
            self._txt_beam_current.isReadOnly() and \
            not self._cb_gun_type.isEnabled() and \
            self._txt_emission_current.isReadOnly() and \
            self._txt_filament_current.isReadOnly() and \
            self._txt_extractor_bias.isReadOnly() and \
            self._txt_beam_diameter.isReadOnly() and \
            self._txt_chamber_pressure.isReadOnly() and \
            self._txt_gun_pressure.isReadOnly() and \
            self._txt_scan_magnification.isReadOnly() and \
            self._txt_working_distance.isReadOnly()

    def hasAcceptableInput(self):
        return _ConditionWidget.hasAcceptableInput(self) and \
            self._txt_beam_voltage.hasAcceptableInput() and \
            self._txt_beam_current.hasAcceptableInput() and \
            self._txt_emission_current.hasAcceptableInput() and \
            self._txt_filament_current.hasAcceptableInput() and \
            self._txt_extractor_bias.hasAcceptableInput() and \
            self._txt_beam_diameter.hasAcceptableInput() and \
            self._txt_chamber_pressure.hasAcceptableInput() and \
            self._txt_gun_pressure.hasAcceptableInput() and \
            self._txt_scan_magnification.hasAcceptableInput() and \
            self._txt_working_distance.hasAcceptableInput()

class ProbeTEMWidget(ProbeEMWidget):

    def __init__(self, parent=None):
        _ConditionWidget.__init__(self, ProbeTEM, parent)

    def _init_ui(self):
        # Controls
        self._cb_lens_mode = QComboBox()
        self._cb_lens_mode.addItems(list(_LENS_MODES))
        self._txt_camera_magnification = NumericalAttributeLineEdit(self.CLASS.camera_magnification)
        self._txt_convergence_angle = NumericalAttributeLineEdit(self.CLASS.convergence_angle)

        # Layouts
        layout = ProbeEMWidget._init_ui(self)
        layout.insertRow(1, '<i>Lens mode</i>', self._cb_lens_mode)
        layout.addRow('Camera magnification', self._txt_camera_magnification)
        layout.addRow('Convergence angle', self._txt_convergence_angle)

        # Signals
        self._cb_lens_mode.currentIndexChanged.connect(self.edited)
        self._txt_camera_magnification.textEdited.connect(self.edited)
        self._txt_convergence_angle.textEdited.connect(self.edited)

        return layout

    def _create_parameter(self):
        return self.CLASS(1e3, LENS_MODE_IMAGE)

    def parameter(self, parameter=None):
        parameter = ProbeEMWidget.parameter(self, parameter)
        parameter.lens_mode = self._cb_lens_mode.currentText()
        parameter.camera_magnification = self._txt_camera_magnification.text()
        parameter.convergence_angle = self._txt_convergence_angle.text()
        return parameter

    def setParameter(self, condition):
        ProbeEMWidget.setParameter(self, condition)
        self._cb_lens_mode.setCurrentIndex(self._cb_lens_mode.findText(condition.lens_mode))
        self._txt_camera_magnification.setText(condition.camera_magnification)
        self._txt_convergence_angle.setText(condition.convergence_angle)

    def setReadOnly(self, state):
        ProbeEMWidget.setReadOnly(self, state)
        self._cb_lens_mode.setEnabled(not state)
        self._txt_camera_magnification.setReadOnly(state)
        self._txt_convergence_angle.setReadOnly(state)

    def isReadOnly(self):
        return ProbeEMWidget.isReadOnly(self) and \
            not self._cb_lens_mode.isEnabled() and \
            self._txt_camera_magnification.isReadOnly() and \
            self._txt_convergence_angle.isReadOnly()

    def hasAcceptableInput(self):
        return ProbeEMWidget.hasAcceptableInput(self) and \
            self._txt_camera_magnification.hasAcceptableInput() and \
            self._txt_convergence_angle.hasAcceptableInput()



"""
Layout utilities
"""

# Standard library modules.

# Third party modules.
from PySide.QtGui import QHBoxLayout, QFormLayout, QLayout

# Local modules.

# Globals and constants variables.

def horizontal(*widgets):
    layout = QHBoxLayout()
    for widget in widgets:
        layout.addWidget(widget)
    return layout

def align_form(layout):
    layouts = []
    _find_layout(layout, QFormLayout, layouts)

    labels = []
    for layout in layouts:
        for index in range(layout.count()):
            layoutitem = layout.itemAt(index, QFormLayout.ItemRole.LabelRole)
            if layoutitem is not None and layoutitem.widget() is not None:
                labels.append(layoutitem.widget())

    maxwidth = float('-inf')
    for label in labels:
        maxwidth = max(maxwidth, label.sizeHint().width())

    for label in labels:
        label.setMinimumSize(maxwidth, 0)

def _find_layout(layout, clasz, layouts):
    for item in map(layout.itemAt, range(layout.count())):
        if isinstance(item, QLayout):
            _find_layout(item, clasz, layouts)
        else:
            if item.widget() is not None and item.widget().layout() is not None:
                _find_layout(item.widget().layout(), clasz, layouts)

    if isinstance(layout, clasz):
        layouts.append(layout)



"""
Fix of QSettings
"""

# Standard library modules.

# Third party modules.
from PySide.QtCore import QSettings

import six

import matplotlib.colors as colors

# Local modules.

# Globals and constants variables.

class Settings(QSettings):

    def valueBool(self, key, default):
        value = self.value(key, default)
        if isinstance(value, six.string_types):
            value = True if value.lower() == 'true' else False
        return value

    def setValueColor(self, key, value):
        r, g, b = colors.colorConverter.to_rgb(value)
        self.setValue(key, ';'.join(map(str, [r, g, b])))

    def valueColor(self, key, default):
        value = self.value(key, default)
        if ';' in value:
            r, g, b = map(float, value.split(';'))
        else:
            r, g, b = colors.colorConverter.to_rgb(value)
        return r, g, b



"""
Base datum widgets
"""

# Standard library modules.

# Third party modules.
from PySide.QtGui import QWidget, QVBoxLayout, QTableView, QHeaderView, QSizePolicy
from PySide.QtCore import Qt

from matplotlib.backends.backend_qt4agg import \
    (FigureCanvasQTAgg as FigureCanvas,
     NavigationToolbar2QT as NavigationToolbar)

# Local modules.

# Globals and constants variables.

class _DatumWidget(QWidget):

    def __init__(self, clasz, controller, datum=None, parent=None):
        QWidget.__init__(self, parent)

        name = clasz.TEMPLATE
        if clasz.CLASS is not None:
            name += ' (%s)' % clasz.CLASS
        self.setAccessibleName(name)

        # Variables
        self._class = clasz
        self._controller = controller

        # Layouts
        layout = QVBoxLayout()
        layout.addLayout(self._init_ui()) # Initialize widgets
        self.setLayout(layout)

        # Defaults
        self.setDatum(datum)

    def _init_ui(self):
        return QVBoxLayout()

    def setDatum(self, datum):
        """
        Sets the datum. Note that ``datum`` could be ``None``.
        """
        pass

    @property
    def CLASS(self):
        return self._class

    @property
    def controller(self):
        return self._controller

class _DatumTableWidget(_DatumWidget):

    def __init__(self, clasz, controller, datum=None, parent=None):
        _DatumWidget.__init__(self, clasz, controller, datum, parent)

    def _init_ui(self):
        # Widgets
        self._table = QTableView()
        self._table.horizontalHeader().setResizeMode(QHeaderView.Stretch)

        # Layouts
        layout = _DatumWidget._init_ui(self)
        layout.addWidget(self._table)

        return layout

    def _create_model(self, datum):
        raise NotImplementedError

    def setDatum(self, datum):
        _DatumWidget.setDatum(self, datum)

        if datum is not None:
            model = self._create_model(datum)
        else:
            model = None
        self._table.setModel(model)

class _DatumFigureWidget(_DatumWidget):

    def __init__(self, clasz, controller, datum=None, parent=None):
        _DatumWidget.__init__(self, clasz, controller, datum, parent)

    def _init_ui(self):
        # Widgets
        figure = self._create_figure()
        self._canvas = FigureCanvas(figure)
        self._canvas.setFocusPolicy(Qt.StrongFocus)
        self._canvas.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self._canvas.updateGeometry()

        self.toolbar = self._create_toolbar(self._canvas)

        # Layouts
        layout = _DatumWidget._init_ui(self)
        layout.addWidget(self._canvas)
        layout.addWidget(self.toolbar)

        return layout

    def _create_figure(self):
        raise NotImplementedError

    def _create_toolbar(self, canvas):
        return NavigationToolbar(canvas, self.parent())

    def _draw_figure(self, fig, datum):
        raise NotImplementedError

    def setDatum(self, datum):
        _DatumWidget.setDatum(self, datum)
        if datum is None:
            self._canvas.figure.clear()
        else:
            self._draw_figure(self._canvas.figure, datum)
        self._canvas.draw()



"""
Composition conditions widgets
"""

# Standard library modules.
from collections import OrderedDict
from operator import methodcaller

# Third party modules.
from PySide.QtGui import \
    (QItemDelegate, QTableView, QToolBar, QMessageBox, QComboBox,
     QLineEdit, QDoubleValidator)
from PySide.QtCore import Qt, QAbstractTableModel, QModelIndex

# Local modules.
from pyhmsa.gui.util.periodictable import PeriodicTableDialog
from pyhmsa.gui.util.icon import getIcon
from pyhmsa.gui.spec.condition.condition import _ConditionWidget

from pyhmsa.spec.condition.composition import CompositionElemental

from pyhmsa.util.element_properties import get_symbol

# Globals and constants variables.
from pyhmsa.spec.condition.composition import _COMPOSITION_UNITS

class _CompositionWidget(_ConditionWidget):

    def _init_ui(self):
        # Widgets
        self._cb_unit = QComboBox()
        self._cb_unit.addItems(list(_COMPOSITION_UNITS))

        # Layouts
        layout = _ConditionWidget._init_ui(self)
        layout.addRow('<i>Unit</i>', self._cb_unit)

        # Signals
        self._cb_unit.currentIndexChanged.connect(self.edited)

        return layout

    def parameter(self, parameter=None):
        parameter = _ConditionWidget.parameter(self, parameter)
        parameter.unit = self._cb_unit.currentText()
        return parameter

    def setParameter(self, condition):
        _ConditionWidget.setParameter(self, condition)
        self._cb_unit.setCurrentIndex(self._cb_unit.findText(condition.unit))

    def setReadOnly(self, state):
        _ConditionWidget.setReadOnly(self, state)
        self._cb_unit.setEnabled(not state)

    def isReadOnly(self):
        return _ConditionWidget.isReadOnly(self) and \
            not self._cb_unit.isEnabled()

class CompositionElementalWidget(_CompositionWidget):

    class _CompositionModel(QAbstractTableModel):

        def __init__(self):
            QAbstractTableModel.__init__(self)
            self.composition = OrderedDict()

        def rowCount(self, *args, **kwargs):
            return len(self.composition)

        def columnCount(self, *args, **kwargs):
            return 2

        def data(self, index, role):
            if not index.isValid() or \
                    not (0 <= index.row() < len(self.composition)):
                return None

            if role == Qt.TextAlignmentRole:
                return Qt.AlignCenter

            if role != Qt.DisplayRole:
                return None

            z, fraction = list(self.composition.items())[index.row()]
            column = index.column()
            if column == 0:
                if z is None:
                    return 'none'
                else:
                    return str(get_symbol(z))
            elif column == 1:
                return str(fraction)

        def headerData(self, section , orientation, role):
            if role != Qt.DisplayRole:
                return None
            if orientation == Qt.Horizontal:
                if section == 0:
                    return 'Element'
                elif section == 1:
                    return 'Fraction'
            elif orientation == Qt.Vertical:
                return str(section + 1)

        def flags(self, index):
            if not index.isValid():
                return Qt.ItemIsEnabled

            return Qt.ItemFlags(QAbstractTableModel.flags(self, index) |
                                Qt.ItemIsEditable)

        def setData(self, index, value, role=Qt.EditRole):
            if not index.isValid() or \
                    not (0 <= index.row() < len(self.composition)):
                return False

            z = list(self.composition.keys())[index.row()]
            column = index.column()
            if column == 0:
                if value in self.composition:
                    return False
                fraction = self.composition.pop(z)
                self.composition[value] = fraction
            elif column == 1:
                self.composition[z] = float(value)

            self.dataChanged.emit(index, index)
            return True

        def insertRows(self, row, count=1, parent=None):
            if count == 0:
                return False
            if parent is None:
                parent = QModelIndex()
            self.beginInsertRows(parent, row, row + count - 1)

            if None in self.composition:
                return False
            self.composition[None] = 0.0

            self.endInsertRows()
            return True

        def removeRows(self, row, count=1, parent=None):
            if count == 0:
                return False
            if parent is None:
                parent = QModelIndex()
            self.beginRemoveRows(parent, row, count + row - 1)

            keys = list(self.composition.keys())
            for key in keys[:row] + keys[row + count:]:
                self.composition.pop(key)

            self.endRemoveRows()
            return True

    class _CompositionDelegate(QItemDelegate):

        def __init__(self, parent=None):
            QItemDelegate.__init__(self, parent)

        def createEditor(self, parent, option, index):
            column = index.column()
            if column == 0:
                editor = PeriodicTableDialog(parent)
                editor.setMultipleSelection(False)
                editor.setRequiresSelection(True)
                return editor
            elif column == 1:
                editor = QLineEdit(parent)
                editor.setValidator(QDoubleValidator())
                return editor
            else:
                return QItemDelegate.createEditor(self, parent, option, index)

        def setEditorData(self, editor, index):
            text = index.model().data(index, Qt.DisplayRole)
            column = index.column()
            if column == 0:
                if text != 'none':
                    editor.setSelection(text)
            elif column == 1:
                editor.setText(text)
            else:
                QItemDelegate.setEditorData(self, editor, index)

        def setModelData(self, editor, model, index):
            column = index.column()
            if column == 0:
                model.setData(index, editor.selection())
            elif column == 1:
                model.setData(index, editor.text())
            else:
                return QItemDelegate.setModelData(self, editor, model, index)

    def __init__(self, parent=None):
        _CompositionWidget.__init__(self, CompositionElemental, parent)

    def _init_ui(self):
        # Widgets
        model = self._CompositionModel()

        self._table = QTableView()
        self._table.setModel(model)
        self._table.setItemDelegate(self._CompositionDelegate(self))
        self._table.horizontalHeader().setStretchLastSection(True)

        self._toolbar = QToolBar()
        action_add = self._toolbar.addAction(getIcon("list-add"), "Add layer")
        action_remove = self._toolbar.addAction(getIcon("list-remove"), "Remove layer")

        # Layouts
        layout = _CompositionWidget._init_ui(self)
        layout.addRow(self._table)
        layout.addRow(self._toolbar)

        # Signals
        action_add.triggered.connect(self._on_add)
        action_remove.triggered.connect(self._on_remove)

        model.dataChanged.connect(self.edited)
        model.rowsInserted.connect(self.edited)
        model.rowsRemoved.connect(self.edited)

        return layout

    def _on_add(self):
        index = self._table.selectionModel().currentIndex()
        model = self._table.model()
        model.insertRows(index.row() + 1)

    def _on_remove(self):
        selection = self._table.selectionModel().selection().indexes()
        if len(selection) == 0:
            QMessageBox.warning(self, "Window layer", "Select a layer")
            return

        model = self._table.model()
        for row in sorted(map(methodcaller('row'), selection), reverse=True):
            model.removeRow(row)

    def _create_parameter(self):
        return self.CLASS('wt%')

    def parameter(self, parameter=None):
        parameter = _CompositionWidget.parameter(self, parameter)
        parameter.update(self._table.model().composition)
        return parameter

    def setParameter(self, condition):
        _CompositionWidget.setParameter(self, condition)
        self._table.model().composition.update(condition)
        self._table.model().reset()

    def setReadOnly(self, state):
        _CompositionWidget.setReadOnly(self, state)
        if state:
            trigger = QTableView.EditTrigger.NoEditTriggers
        else:
            trigger = QTableView.EditTrigger.AllEditTriggers
        self._table.setEditTriggers(trigger)
        self._toolbar.setEnabled(not state)

    def isReadOnly(self):
        return _CompositionWidget.isReadOnly(self) and \
            self._table.editTriggers() == QTableView.EditTrigger.NoEditTriggers and \
            not self._toolbar.isEnabled()



"""
Acquisition conditions widgets
"""

# Standard library modules.

# Third party modules.
from PySide.QtGui import QComboBox

# Local modules.
from pyhmsa.gui.util.parameter import NumericalAttributeLineEdit
from pyhmsa.gui.spec.condition.condition import _ConditionWidget
from pyhmsa.gui.spec.condition.specimenposition import \
    SpecimenPositionWidget, SpecimenPositionListWidget

from pyhmsa.spec.condition.acquisition import \
    (AcquisitionPoint, AcquisitionMultipoint, AcquisitionRasterLinescan,
     AcquisitionRasterXY, AcquisitionRasterXYZ)
from pyhmsa.spec.condition.specimenposition import SpecimenPosition

# Globals and constants variables.
from pyhmsa.spec.condition.acquisition import \
    _RASTER_MODES, _POSITION_LOCATIONS, _RASTER_MODES_Z

class _AcquisitionWidget(_ConditionWidget):

    def _init_ui(self):
        # Widgets
        self._txt_dwell_time = NumericalAttributeLineEdit(self.CLASS.dwell_time)
        self._txt_total_time = NumericalAttributeLineEdit(self.CLASS.total_time)
        self._txt_dwell_time_live = NumericalAttributeLineEdit(self.CLASS.dwell_time_live)

        # Layouts
        layout = _ConditionWidget._init_ui(self)
        layout.addRow('Dwell time', self._txt_dwell_time)
        layout.addRow('Total time', self._txt_total_time)
        layout.addRow('Dwell time (live)', self._txt_dwell_time_live)

        # Signals
        self._txt_dwell_time.textEdited.connect(self.edited)
        self._txt_total_time.textEdited.connect(self.edited)
        self._txt_dwell_time_live.textEdited.connect(self.edited)

        return layout

    def parameter(self, parameter=None):
        parameter = _ConditionWidget.parameter(self, parameter)
        parameter.dwell_time = self._txt_dwell_time.text()
        parameter.total_time = self._txt_total_time.text()
        parameter.dwell_time_live = self._txt_dwell_time_live.text()
        return parameter

    def setParameter(self, condition):
        _ConditionWidget.setParameter(self, condition)
        self._txt_dwell_time.setText(condition.dwell_time)
        self._txt_total_time.setText(condition.total_time)
        self._txt_dwell_time_live.setText(condition.dwell_time_live)

    def setReadOnly(self, state):
        _ConditionWidget.setReadOnly(self, state)
        self._txt_dwell_time.setReadOnly(state)
        self._txt_total_time.setReadOnly(state)
        self._txt_dwell_time_live.setReadOnly(state)

    def isReadOnly(self):
        return _ConditionWidget.isReadOnly(self) and \
            self._txt_dwell_time.isReadOnly() and \
            self._txt_total_time.isReadOnly() and \
            self._txt_dwell_time_live.isReadOnly()

    def hasAcceptableInput(self):
        return _ConditionWidget.hasAcceptableInput(self) and \
            self._txt_dwell_time.hasAcceptableInput() and \
            self._txt_total_time.hasAcceptableInput() and \
            self._txt_dwell_time_live.hasAcceptableInput()

class AcquisitionPointWidget(_AcquisitionWidget):

    def __init__(self, parent=None):
        _AcquisitionWidget.__init__(self, AcquisitionPoint, parent)

    def _init_ui(self):
        # Widgets
        self._wdg_position = SpecimenPositionWidget(inline=True)

        # Layouts
        layout = _AcquisitionWidget._init_ui(self)
        layout.insertRow(0, 'Position', self._wdg_position)

        # Signals
        self._wdg_position.edited.connect(self.edited)

        return layout

    def _create_parameter(self):
        return self.CLASS(position=SpecimenPosition())

    def parameter(self, parameter=None):
        parameter = _AcquisitionWidget.parameter(self, parameter)
        parameter.position = self._wdg_position.condition()
        return parameter

    def setParameter(self, condition):
        _AcquisitionWidget.setParameter(self, condition)
        if condition.position is not None:
            self._wdg_position.setParameter(condition.position)

    def setReadOnly(self, state):
        _AcquisitionWidget.setReadOnly(self, state)
        self._wdg_position.setReadOnly(state)

    def isReadOnly(self):
        return _AcquisitionWidget.isReadOnly(self) and \
            self._wdg_position.isReadOnly()

    def hasAcceptableInput(self):
        return _AcquisitionWidget.hasAcceptableInput(self) and \
            self._wdg_position.hasAcceptableInput()

class AcquisitionMultipointWidget(_AcquisitionWidget):

    def __init__(self, parent=None):
        _AcquisitionWidget.__init__(self, AcquisitionMultipoint, parent)

    def _init_ui(self):
        # Widgets
        self._wdg_positions = SpecimenPositionListWidget()

        # Layouts
        layout = _AcquisitionWidget._init_ui(self)
        layout.addRow('Positions', self._wdg_positions)

        # Signals
        self._wdg_positions.edited.connect(self.edited)

        return layout

    def parameter(self, parameter=None):
        parameter = _AcquisitionWidget.parameter(self, parameter)
        parameter.positions.clear()
        parameter.positions.extend(self._wdg_positions.positions())
        return parameter

    def setParameter(self, condition):
        _AcquisitionWidget.setParameter(self, condition)
        self._wdg_positions.setPositions(condition.positions)

    def setReadOnly(self, state):
        _AcquisitionWidget.setReadOnly(self, state)
        self._wdg_positions.setReadOnly(state)

    def isReadOnly(self):
        return _AcquisitionWidget.isReadOnly(self) and \
            self._wdg_positions.isReadOnly()

    def hasAcceptableInput(self):
        return _AcquisitionWidget.hasAcceptableInput(self) and \
            self._wdg_positions.hasAcceptableInput()

class _AcquisitionRasterWidget(_AcquisitionWidget):

    def _init_ui(self):
        # Widgets
        self._cb_raster_mode = QComboBox()
        self._cb_raster_mode.addItems([None] + list(_RASTER_MODES))

        # Layouts
        layout = _AcquisitionWidget._init_ui(self)
        layout.addRow('Raster mode', self._cb_raster_mode)

        # Singals
        self._cb_raster_mode.currentIndexChanged.connect(self.edited)

        return layout

    def parameter(self, parameter=None):
        parameter = _AcquisitionWidget.parameter(self, parameter)
        parameter.raster_mode = self._cb_raster_mode.currentText()
        return parameter

    def setParameter(self, condition):
        _AcquisitionWidget.setParameter(self, condition)
        self._cb_raster_mode.setCurrentIndex(self._cb_raster_mode.findText(condition.raster_mode))

    def setReadOnly(self, state):
        _AcquisitionWidget.setReadOnly(self, state)
        self._cb_raster_mode.setEnabled(not state)

    def isReadOnly(self):
        return _AcquisitionWidget.isReadOnly(self) and \
            not self._cb_raster_mode.isEnabled()

class AcquisitionRasterLinescanWidget(_AcquisitionRasterWidget):

    def __init__(self, parent=None):
        _AcquisitionRasterWidget.__init__(self, AcquisitionRasterLinescan, parent)

    def _init_ui(self):
        # Widgets
        self._txt_step_count = NumericalAttributeLineEdit(self.CLASS.step_count)
        self._txt_step_count.setFormat('{0:d}')
        self._txt_step_size = NumericalAttributeLineEdit(self.CLASS.step_size)
        self._txt_frame_count = NumericalAttributeLineEdit(self.CLASS.frame_count)
        self._txt_frame_count.setFormat('{0:d}')
        self._wdg_position_start = SpecimenPositionWidget(inline=True)
        self._wdg_position_end = SpecimenPositionWidget(inline=True)

        # Layout
        layout = _AcquisitionRasterWidget._init_ui(self)
        layout.insertRow(0, '<i>Step count</i>', self._txt_step_count)
        layout.insertRow(1, 'Step size', self._txt_step_size)
        layout.addRow('Frame count', self._txt_frame_count)
        layout.addRow('Start position', self._wdg_position_start)
        layout.addRow('End position', self._wdg_position_end)

        # Signals
        self._txt_step_count.textEdited.connect(self.edited)
        self._txt_step_size.textEdited.connect(self.edited)
        self._txt_frame_count.textEdited.connect(self.edited)
        self._wdg_position_start.edited.connect(self.edited)
        self._wdg_position_end.edited.connect(self.edited)

        return layout

    def _create_parameter(self):
        return self.CLASS(1)

    def parameter(self, parameter=None):
        parameter = _AcquisitionRasterWidget.parameter(self, parameter)
        parameter.step_count = self._txt_step_count.text()
        parameter.step_size = self._txt_step_size.text()
        parameter.frame_count = self._txt_frame_count.text()
        parameter.position_start = self._wdg_position_start.condition()
        parameter.position_end = self._wdg_position_end.condition()
        return parameter

    def setParameter(self, condition):
        _AcquisitionRasterWidget.setParameter(self, condition)
        self._txt_step_count.setText(condition.step_count)
        self._txt_step_size.setText(condition.step_size)
        self._txt_frame_count.setText(condition.frame_count)
        self._wdg_position_start.setParameter(condition.position_start)
        self._wdg_position_end.setParameter(condition.position_end)

    def setReadOnly(self, state):
        _AcquisitionRasterWidget.setReadOnly(self, state)
        self._txt_step_count.setReadOnly(state)
        self._txt_step_size.setReadOnly(state)
        self._txt_frame_count.setReadOnly(state)
        self._wdg_position_start.setReadOnly(state)
        self._wdg_position_end.setReadOnly(state)

    def isReadOnly(self):
        return _AcquisitionRasterWidget.isReadOnly(self) and \
            self._txt_step_count.isReadOnly() and \
            self._txt_step_size.isReadOnly() and \
            self._txt_frame_count.isReadOnly() and \
            self._wdg_position_start.isReadOnly() and \
            self._wdg_position_end.isReadOnly()

    def hasAcceptableInput(self):
        return _AcquisitionRasterWidget.hasAcceptableInput(self) and \
            self._txt_step_count.hasAcceptableInput() and \
            self._txt_step_size.hasAcceptableInput() and \
            self._txt_frame_count.hasAcceptableInput() and \
            self._wdg_position_start.hasAcceptableInput() and \
            self._wdg_position_end.hasAcceptableInput()

class AcquisitionRasterXYWidget(_AcquisitionRasterWidget):

    def __init__(self, parent=None):
        _AcquisitionRasterWidget.__init__(self, AcquisitionRasterXY, parent)

    def _init_ui(self):
        # Widgets
        self._txt_step_count_x = NumericalAttributeLineEdit(self.CLASS.step_count_x)
        self._txt_step_count_x.setFormat('{0:d}')
        self._txt_step_count_y = NumericalAttributeLineEdit(self.CLASS.step_count_y)
        self._txt_step_count_y.setFormat('{0:d}')
        self._txt_step_size_x = NumericalAttributeLineEdit(self.CLASS.step_size_x)
        self._txt_step_size_y = NumericalAttributeLineEdit(self.CLASS.step_size_y)
        self._txt_frame_count = NumericalAttributeLineEdit(self.CLASS.frame_count)
        self._txt_frame_count.setFormat('{0:d}')
        self._cb_location = QComboBox()
        self._cb_location.addItems(list(_POSITION_LOCATIONS))
        self._wdg_position = SpecimenPositionWidget(inline=True)

        # Layouts
        layout = _AcquisitionRasterWidget._init_ui(self)
        layout.insertRow(0, '<i>Step count (x)</i>', self._txt_step_count_x)
        layout.insertRow(1, '<i>Step count (y)</i>', self._txt_step_count_y)
        layout.insertRow(2, 'Step size (x)', self._txt_step_size_x)
        layout.insertRow(3, 'Step size (y)', self._txt_step_size_y)
        layout.addRow('Frame count', self._txt_frame_count)
        layout.addRow('Position', self._cb_location)
        layout.addWidget(self._wdg_position)

        # Signals
        self._txt_step_count_x.textEdited.connect(self.edited)
        self._txt_step_count_y.textEdited.connect(self.edited)
        self._txt_step_size_x.textEdited.connect(self.edited)
        self._txt_step_size_y.textEdited.connect(self.edited)
        self._txt_frame_count.textEdited.connect(self.edited)
        self._cb_location.currentIndexChanged.connect(self.edited)
        self._wdg_position.edited.connect(self.edited)

        return layout

    def _create_parameter(self):
        return self.CLASS(1, 1)

    def parameter(self, parameter=None):
        parameter = _AcquisitionRasterWidget.parameter(self, parameter)
        parameter.step_count_x = self._txt_step_count_x.text()
        parameter.step_count_y = self._txt_step_count_y.text()
        parameter.step_size_x = self._txt_step_size_x.text()
        parameter.step_size_y = self._txt_step_size_y.text()
        parameter.frame_count = self._txt_frame_count.text()
        parameter.position = (self._wdg_position.condition(),
                              self._cb_location.currentText())
        return parameter

    def setParameter(self, condition):
        _AcquisitionRasterWidget.setParameter(self, condition)
        self._txt_step_count_x.setText(condition.step_count_x)
        self._txt_step_count_y.setText(condition.step_count_y)
        self._txt_step_size_x.setText(condition.step_size_x)
        self._txt_step_size_y.setText(condition.step_size_y)
        self._txt_frame_count.setText(condition.frame_count)

        position, location = condition.get_position(True)
        self._cb_location.setCurrentIndex(self._cb_location.findText(location))
        if position is not None:
            self._wdg_position.setParameter(position)

    def setReadOnly(self, state):
        _AcquisitionRasterWidget.setReadOnly(self, state)
        self._txt_step_count_x.setReadOnly(state)
        self._txt_step_count_y.setReadOnly(state)
        self._txt_step_size_x.setReadOnly(state)
        self._txt_step_size_y.setReadOnly(state)
        self._txt_frame_count.setReadOnly(state)
        self._cb_location.setEnabled(not state)
        self._wdg_position.setReadOnly(state)

    def isReadOnly(self):
        return _AcquisitionRasterWidget.isReadOnly(self) and \
            self._txt_step_count_x.isReadOnly() and \
            self._txt_step_count_y.isReadOnly() and \
            self._txt_step_size_x.isReadOnly() and \
            self._txt_step_size_y.isReadOnly() and \
            self._txt_frame_count.isReadOnly() and \
            not self._cb_location.isEnabled() and \
            self._wdg_position.isReadOnly()

    def hasAcceptableInput(self):
        return _AcquisitionRasterWidget.hasAcceptableInput(self) and \
            self._txt_step_count_x.hasAcceptableInput() and \
            self._txt_step_count_y.hasAcceptableInput() and \
            self._txt_step_size_x.hasAcceptableInput() and \
            self._txt_step_size_y.hasAcceptableInput() and \
            self._txt_frame_count.hasAcceptableInput() and \
            self._wdg_position.hasAcceptableInput()

class AcquisitionRasterXYZWidget(_AcquisitionRasterWidget):

    def __init__(self, parent=None):
        _AcquisitionRasterWidget.__init__(self, AcquisitionRasterXYZ, parent)

    def _init_ui(self):
        # Widgets
        self._txt_step_count_x = NumericalAttributeLineEdit(self.CLASS.step_count_x)
        self._txt_step_count_x.setFormat('{0:d}')
        self._txt_step_count_y = NumericalAttributeLineEdit(self.CLASS.step_count_y)
        self._txt_step_count_y.setFormat('{0:d}')
        self._txt_step_count_z = NumericalAttributeLineEdit(self.CLASS.step_count_z)
        self._txt_step_count_z.setFormat('{0:d}')
        self._txt_step_size_x = NumericalAttributeLineEdit(self.CLASS.step_size_x)
        self._txt_step_size_y = NumericalAttributeLineEdit(self.CLASS.step_size_y)
        self._txt_step_size_z = NumericalAttributeLineEdit(self.CLASS.step_size_z)
        self._cb_raster_mode_z = QComboBox()
        self._cb_raster_mode_z.addItems([None] + list(_RASTER_MODES_Z))
        self._cb_location = QComboBox()
        self._cb_location.addItems(list(_POSITION_LOCATIONS))
        self._wdg_position = SpecimenPositionWidget(inline=True)

        # Layouts
        layout = _AcquisitionRasterWidget._init_ui(self)
        layout.insertRow(0, '<i>Step count (x)</i>', self._txt_step_count_x)
        layout.insertRow(1, '<i>Step count (y)</i>', self._txt_step_count_y)
        layout.insertRow(2, '<i>Step count (z)</i>', self._txt_step_count_z)
        layout.insertRow(3, 'Step size (x)', self._txt_step_size_x)
        layout.insertRow(4, 'Step size (y)', self._txt_step_size_y)
        layout.insertRow(5, 'Step size (z)', self._txt_step_size_z)
        layout.addRow('Raster mode (z)', self._cb_raster_mode_z)
        layout.addRow('Position', self._cb_location)
        layout.addWidget(self._wdg_position)

        # Signals
        self._txt_step_count_x.textEdited.connect(self.edited)
        self._txt_step_count_y.textEdited.connect(self.edited)
        self._txt_step_count_z.textEdited.connect(self.edited)
        self._txt_step_size_x.textEdited.connect(self.edited)
        self._txt_step_size_y.textEdited.connect(self.edited)
        self._txt_step_size_z.textEdited.connect(self.edited)
        self._cb_raster_mode_z.currentIndexChanged.connect(self.edited)
        self._cb_location.currentIndexChanged.connect(self.edited)
        self._wdg_position.edited.connect(self.edited)

        return layout

    def _create_parameter(self):
        return self.CLASS(1, 1, 1)

    def parameter(self, parameter=None):
        parameter = _AcquisitionRasterWidget.parameter(self, parameter)
        parameter.step_count_x = self._txt_step_count_x.text()
        parameter.step_count_y = self._txt_step_count_y.text()
        parameter.step_count_z = self._txt_step_count_z.text()
        parameter.step_size_x = self._txt_step_size_x.text()
        parameter.step_size_y = self._txt_step_size_y.text()
        parameter.step_size_z = self._txt_step_size_z.text()
        parameter.raster_mode_z = self._cb_raster_mode_z.currentText()
        parameter.position = (self._wdg_position.condition(),
                              self._cb_location.currentText())
        return parameter

    def setParameter(self, condition):
        _AcquisitionRasterWidget.setParameter(self, condition)
        self._txt_step_count_x.setText(condition.step_count_x)
        self._txt_step_count_y.setText(condition.step_count_y)
        self._txt_step_count_z.setText(condition.step_count_z)
        self._txt_step_size_x.setText(condition.step_size_x)
        self._txt_step_size_y.setText(condition.step_size_y)
        self._txt_step_size_z.setText(condition.step_size_z)
        self._cb_raster_mode_z.setCurrentIndex(self._cb_raster_mode_z.findText(condition.raster_mode_z))

        position, location = condition.get_position(True)
        self._cb_location.setCurrentIndex(self._cb_location.findText(location))
        if position is not None:
            self._wdg_position.setParameter(position)

    def setReadOnly(self, state):
        _AcquisitionRasterWidget.setReadOnly(self, state)
        self._txt_step_count_x.setReadOnly(state)
        self._txt_step_count_y.setReadOnly(state)
        self._txt_step_count_z.setReadOnly(state)
        self._txt_step_size_x.setReadOnly(state)
        self._txt_step_size_y.setReadOnly(state)
        self._txt_step_size_z.setReadOnly(state)
        self._cb_raster_mode_z.setEnabled(not state)
        self._cb_location.setEnabled(not state)
        self._wdg_position.setReadOnly(state)

    def isReadOnly(self):
        return _AcquisitionRasterWidget.isReadOnly(self) and \
            self._txt_step_count_x.isReadOnly() and \
            self._txt_step_count_y.isReadOnly() and \
            self._txt_step_count_z.isReadOnly() and \
            self._txt_step_size_x.isReadOnly() and \
            self._txt_step_size_y.isReadOnly() and \
            self._txt_step_size_z.isReadOnly() and \
            not self._cb_raster_mode_z.isEnabled() and \
            not self._cb_location.isEnabled() and \
            self._wdg_position.isReadOnly()

    def hasAcceptableInput(self):
        return _AcquisitionRasterWidget.hasAcceptableInput(self) and \
            self._txt_step_count_x.hasAcceptableInput() and \
            self._txt_step_count_y.hasAcceptableInput() and \
            self._txt_step_count_z.hasAcceptableInput() and \
            self._txt_step_size_x.hasAcceptableInput() and \
            self._txt_step_size_y.hasAcceptableInput() and \
            self._txt_step_size_z.hasAcceptableInput() and \
            self._wdg_position.hasAcceptableInput()




"""
Header widget
"""

# Standard library modules.

# Third party modules.
from PySide.QtGui import QDateEdit, QTimeEdit
from PySide.QtCore import QDate, QTime

# Local modules.
from pyhmsa.gui.util.parameter import ParameterWidget, TextAttributeLineEdit

from pyhmsa.spec.header import Header

# Globals and constants variables.

class HeaderWidget(ParameterWidget):

    def __init__(self, parent=None):
        ParameterWidget.__init__(self, Header, parent)

    def _init_ui(self):
        # Widgets
        self._txt_title = TextAttributeLineEdit(self.CLASS.title)
        self._txt_author = TextAttributeLineEdit(self.CLASS.author)
        self._txt_owner = TextAttributeLineEdit(self.CLASS.owner)
        self._txt_date = QDateEdit()
        self._txt_date.setDisplayFormat('yyyy-MM-dd')
        self._txt_time = QTimeEdit()
        self._txt_time.setDisplayFormat('hh:mm:ss')
        self._txt_timezone = TextAttributeLineEdit(self.CLASS.timezone)
        self._txt_checksum = TextAttributeLineEdit(self.CLASS.checksum)
        self._txt_checksum.setReadOnly(True)

        # Layouts
        layout = ParameterWidget._init_ui(self)
        layout.addRow('Title', self._txt_title)
        layout.addRow('Author', self._txt_author)
        layout.addRow('Owner', self._txt_owner)
        layout.addRow('Date', self._txt_date)
        layout.addRow('Time', self._txt_time)
        layout.addRow('Timezone', self._txt_timezone)
        layout.addRow('Checksum', self._txt_checksum)

        # Signals
        self._txt_title.textEdited.connect(self.edited)
        self._txt_author.textEdited.connect(self.edited)
        self._txt_owner.textEdited.connect(self.edited)
        self._txt_date.dateChanged.connect(self.edited)
        self._txt_time.timeChanged.connect(self.edited)
        self._txt_timezone.textEdited.connect(self.edited)
        self._txt_checksum.textEdited.connect(self.edited)

        return layout

    def parameter(self, parameter=None):
        parameter = ParameterWidget.parameter(self, parameter)
        parameter.title = self._txt_title.text()
        parameter.author = self._txt_author.text()
        parameter.owner = self._txt_owner.text()
        parameter.date = self._txt_date.date().toString('yyyy-MM-dd')
        parameter.time = self._txt_time.time().toString('hh:mm:ss')
        parameter.timezone = self._txt_timezone.text()
        return parameter

    def setParameter(self, header):
        ParameterWidget.setParameter(self, header)
        self._txt_title.setText(header.title)
        self._txt_author.setText(header.author)
        self._txt_owner.setText(header.owner)
        date = header.date
        if date is not None:
            self._txt_date.setDate(QDate(date.year, date.month, date.day))
        time = header.time
        if time is not None:
            self._txt_time.setTime(QTime(time.hour, time.minute, time.second))
        self._txt_timezone.setText(header.timezone)
        checksum = header.checksum
        if checksum is not None:
            self._txt_checksum.setText(checksum.value)

    def header(self, header=None):
        return self.parameter(header)

    def setHeader(self, header):
        self.setParameter(header)

    def setReadOnly(self, state):
        ParameterWidget.setReadOnly(self, state)
        self._txt_title.setReadOnly(state)
        self._txt_author.setReadOnly(state)
        self._txt_owner.setReadOnly(state)
        self._txt_date.setReadOnly(state)
        self._txt_time.setReadOnly(state)
        self._txt_timezone.setReadOnly(state)
        self._txt_checksum.setReadOnly(state)

    def isReadOnly(self):
        return ParameterWidget.isReadOnly(self) and \
            self._txt_title.isReadOnly() and \
            self._txt_author.isReadOnly() and \
            self._txt_owner.isReadOnly() and \
            self._txt_date.isReadOnly() and \
            self._txt_time.isReadOnly() and \
            self._txt_timezone.isReadOnly() and \
            self._txt_checksum.isReadOnly()

    def hasAcceptableInput(self):
        return ParameterWidget.hasAcceptableInput(self) and \
            self._txt_title.hasAcceptableInput() and \
            self._txt_author.hasAcceptableInput() and \
            self._txt_owner.hasAcceptableInput() and \
            self._txt_date.hasAcceptableInput() and \
            self._txt_time.hasAcceptableInput() and \
            self._txt_timezone.hasAcceptableInput() and \
            self._txt_checksum.hasAcceptableInput()



"""
Calibration widgets
"""

# Standard library modules.
import re

# Third party modules.
from PySide.QtGui import QValidator, QComboBox, QStackedWidget

from pkg_resources import iter_entry_points #@UnresolvedImport

# Local modules.
from pyhmsa.gui.util.parameter import \
    ParameterWidget, _AttributeLineEdit, TextAttributeLineEdit, UnitAttributeLineEdit, NumericalAttributeLineEdit

from pyhmsa.spec.condition.calibration import \
    (_Calibration, CalibrationConstant, CalibrationLinear,
     CalibrationPolynomial, CalibrationExplicit)

# Globals and constants variables.

class _CalibrationWidget(ParameterWidget):

    def _init_ui(self):
        # Widgets
        self._txt_quantity = TextAttributeLineEdit(self.CLASS.quantity)
        self._txt_unit = UnitAttributeLineEdit(self.CLASS.unit)

        # Layouts
        layout = ParameterWidget._init_ui(self)
        layout.addRow('<i>Quantity</i>', self._txt_quantity)
        layout.addRow('<i>Unit</i>', self._txt_unit)

        # Signals
        self._txt_quantity.textEdited.connect(self.edited)
        self._txt_unit.textEdited.connect(self.edited)

        return layout

    def parameter(self, parameter=None):
        parameter = ParameterWidget.parameter(self, parameter)
        parameter.quantity = self._txt_quantity.text()
        parameter.unit = self._txt_unit.text()
        return parameter

    def setParameter(self, calibration):
        self._txt_quantity.setText(calibration.quantity)
        self._txt_unit.setText(calibration.unit)

    def setReadOnly(self, state):
        ParameterWidget.setReadOnly(self, state)
        self._txt_quantity.setReadOnly(state)
        self._txt_unit.setReadOnly(state)

    def isReadOnly(self):
        return ParameterWidget.isReadOnly(self) and \
            self._txt_quantity.isReadOnly() and \
            self._txt_unit.isReadOnly()

    def hasAcceptableInput(self):
        return ParameterWidget.hasAcceptableInput(self) and \
            self._txt_quantity.hasAcceptableInput() and \
            self._txt_unit.hasAcceptableInput()

    def calibration(self):
        return self.parameter()

    def setCalibration(self, calibration):
        self.setParameter(calibration)

class CalibrationConstantWidget(_CalibrationWidget):

    def __init__(self, parent=None):
        _CalibrationWidget.__init__(self, CalibrationConstant, parent)
        self.setAccessibleName('constant')

    def _init_ui(self):
        # Widgets
        self._txt_value = NumericalAttributeLineEdit(self.CLASS.value)

        # Layouts
        layout = _CalibrationWidget._init_ui(self)
        layout.addRow('<i>Value</i>', self._txt_value)

        # Signals
        self._txt_value.textEdited.connect(self.edited)

        return layout

    def _create_parameter(self):
        return self.CLASS('Quantity', 'm', 0.0)

    def parameter(self, parameter=None):
        parameter = _CalibrationWidget.parameter(self, parameter)
        parameter.value = self._txt_value.text()
        return parameter

    def setParameter(self, calibration):
        _CalibrationWidget.setParameter(self, calibration)
        self._txt_value.setText(calibration.value)

    def setReadOnly(self, state):
        _CalibrationWidget.setReadOnly(self, state)
        self._txt_value.setReadOnly(state)

    def isReadOnly(self):
        return _CalibrationWidget.isReadOnly(self) and \
            self._txt_value.isReadOnly()

    def hasAcceptableInput(self):
        return _CalibrationWidget.hasAcceptableInput(self) and \
            self._txt_value.hasAcceptableInput()

class CalibrationLinearWidget(_CalibrationWidget):

    def __init__(self, parent=None):
        _CalibrationWidget.__init__(self, CalibrationLinear, parent)
        self.setAccessibleName('linear')

    def _init_ui(self):
        # Widgets
        self._txt_gain = NumericalAttributeLineEdit(self.CLASS.gain)
        self._txt_offset = NumericalAttributeLineEdit(self.CLASS.offset)

        # Layouts
        layout = _CalibrationWidget._init_ui(self)
        layout.addRow('<i>Gain</i>', self._txt_gain)
        layout.addRow('<i>Offset</i>', self._txt_offset)

        # Signals
        self._txt_gain.textEdited.connect(self.edited)
        self._txt_offset.textEdited.connect(self.edited)

        return layout

    def _create_parameter(self):
        return self.CLASS('Quantity', 'm', 0.0, 0.0)

    def parameter(self, parameter=None):
        parameter = _CalibrationWidget.parameter(self, parameter)
        parameter.gain = self._txt_gain.text()
        parameter.offset = self._txt_offset.text()
        return parameter

    def setParameter(self, calibration):
        _CalibrationWidget.setParameter(self, calibration)
        self._txt_gain.setText(calibration.gain)
        self._txt_offset.setText(calibration.offset)

    def setReadOnly(self, state):
        _CalibrationWidget.setReadOnly(self, state)
        self._txt_gain.setReadOnly(state)
        self._txt_offset.setReadOnly(state)

    def isReadOnly(self):
        return _CalibrationWidget.isReadOnly(self) and \
            self._txt_gain.isReadOnly() and \
            self._txt_offset.isReadOnly()

    def hasAcceptableInput(self):
        return _CalibrationWidget.hasAcceptableInput(self) and \
            self._txt_gain.hasAcceptableInput() and \
            self._txt_offset.hasAcceptableInput()

class _CoefficientsLineEdit(_AttributeLineEdit):

    _PATTERN = re.compile(r'^(?P<coef>[\d\.]*)\s*(?P<x>[x]?)(?:[\^](?P<exp>\d*))?$')

    @staticmethod
    def parse_coefficients(text):
        # Parse terms
        terms = {}
        for term in text.split('+'):
            term = term.strip()
            match = _CoefficientsLineEdit._PATTERN.match(term)
            if not match:
                raise ValueError('Unparseable term: %s' % term)

            matchdict = match.groupdict()
            coefficient = float(matchdict['coef'] or (1.0 if matchdict['x'] else 0.0))
            exponent = int(matchdict['exp'] or 1 if matchdict['x'] else 0)

            terms.setdefault(exponent, 0.0)
            terms[exponent] += coefficient

        # Add missing terms
        for exponent in range(max(terms.keys()) + 1):
            if exponent not in terms:
                terms[exponent] = 0.0

        coefficients = []
        for exponent in sorted(terms.keys(), reverse=True):
            coefficients.append(terms[exponent])

        return coefficients

    @staticmethod
    def write_coefficients(coefficients):
        terms = []

        for order, coefficient in enumerate(reversed(coefficients)):
            if coefficient == 0.0: continue
            coefficient_text = str(coefficient) if coefficient != 1.0 else ''
            if order == 0:
                terms.append(coefficient_text)
            elif order == 1:
                terms.append(coefficient_text + "x")
            else:
                terms.append(coefficient_text + "x^" + str(order))

        return ' + '.join(reversed(terms))

    class _Validator(QValidator):

        def validate(self, text, pos):
            try:
                _CoefficientsLineEdit.parse_coefficients(text)
            except:
                return QValidator.Intermediate
            else:
                return QValidator.Acceptable

        def fixup(self, text):
            return text

    def __init__(self, attribute, *args, **kwargs):
        _AttributeLineEdit.__init__(self, attribute, *args, **kwargs)

        self.setValidator(self._Validator())

        self.editingFinished.connect(self._on_editing_finished)

    def _on_editing_finished(self):
        if not self.hasAcceptableInput():
            return
        self.setText(self.text())

    def setText(self, text):
        if text is None:
            text = ''
        else:
            text = _CoefficientsLineEdit.write_coefficients(text)
        return _AttributeLineEdit.setText(self, text)

    def text(self):
        if not self.hasAcceptableInput():
            raise ValueError('Invalid text')

        text = _AttributeLineEdit.text(self)
        if len(text.strip()) == 0:
            return None

        return _CoefficientsLineEdit.parse_coefficients(text)

class CalibrationPolynomialWidget(_CalibrationWidget):

    def __init__(self, parent=None):
        _CalibrationWidget.__init__(self, CalibrationPolynomial, parent)
        self.setAccessibleName('polynomial')

    def _init_ui(self):
        # Widgets
        self._txt_coefficients = _CoefficientsLineEdit(self.CLASS.coefficients)

        # Layout
        layout = _CalibrationWidget._init_ui(self)
        layout.addRow('<i>Coefficients</i>', self._txt_coefficients)

        # Signals
        self._txt_coefficients.textEdited.connect(self.edited)

        return layout

    def _create_parameter(self):
        return self.CLASS('Quantity', 'm', (0,))

    def parameter(self, parameter=None):
        parameter = _CalibrationWidget.parameter(self, parameter)
        parameter.coefficients = self._txt_coefficients.text()
        return parameter

    def setParameter(self, calibration):
        _CalibrationWidget.setParameter(self, calibration)
        self._txt_coefficients.setText(calibration.coefficients)

    def setReadOnly(self, state):
        _CalibrationWidget.setReadOnly(self, state)
        self._txt_coefficients.setReadOnly(state)

    def isReadOnly(self):
        return _CalibrationWidget.isReadOnly(self) and \
            self._txt_coefficients.isReadOnly()

    def hasAcceptableInput(self):
        return _CalibrationWidget.hasAcceptableInput(self) and \
            self._txt_coefficients.hasAcceptableInput()

class _ExplicitLineEdit(_AttributeLineEdit):

    _PATTERN = re.compile(r'[,;]')

    class _Validator(QValidator):

        def validate(self, text, pos):
            try:
                map(float, _ExplicitLineEdit._PATTERN.split(text))
            except:
                return QValidator.Intermediate
            else:
                return QValidator.Acceptable

        def fixup(self, text):
            return text

    def __init__(self, attribute, *args, **kwargs):
        _AttributeLineEdit.__init__(self, attribute, *args, **kwargs)

        self.setValidator(self._Validator())

        self.editingFinished.connect(self._on_editing_finished)

    def _on_editing_finished(self):
        if not self.hasAcceptableInput():
            return
        self.setText(self.text())

    def setText(self, text):
        if text is None:
            text = ''
        else:
            text = ','.join(map(str, text))
        return _AttributeLineEdit.setText(self, text)

    def text(self):
        if not self.hasAcceptableInput():
            raise ValueError('Invalid text')

        text = _AttributeLineEdit.text(self)
        if len(text.strip()) == 0:
            return None

        return self._PATTERN.split(text)

class _ValuesExplicitWidget(_ExplicitLineEdit):

    def text(self):
        return list(map(float, _ExplicitLineEdit.text(self)))

class _LabelsExplicitWidget(_ExplicitLineEdit):
    pass

class CalibrationExplicitWidget(_CalibrationWidget):

    def __init__(self, parent=None):
        _CalibrationWidget.__init__(self, CalibrationExplicit, parent=parent)
        self.setAccessibleName('explicit')

    def _init_ui(self):
        # Widgets
        self._txt_values = _ValuesExplicitWidget(self.CLASS.values)
        self._txt_labels = _LabelsExplicitWidget(self.CLASS.labels)

        # Layouts
        layout = _CalibrationWidget._init_ui(self)
        layout.addRow('Values', self._txt_values)
        layout.addRow('Labels', self._txt_labels)

        # Signals
        self._txt_values.textEdited.connect(self.edited)
        self._txt_labels.textEdited.connect(self.edited)

        return layout

    def _create_parameter(self):
        return self.CLASS('Quantity', 'm', (0,))

    def parameter(self, parameter=None):
        parameter = _CalibrationWidget.parameter(self, parameter)
        parameter.values = self._txt_values.text()
        parameter.labels = self._txt_labels.text()
        return parameter

    def setParameter(self, calibration):
        _CalibrationWidget.setParameter(self, calibration)
        self._txt_values.setText(calibration.values)
        self._txt_labels.setText(calibration.labels)

    def setReadOnly(self, state):
        _CalibrationWidget.setReadOnly(self, state)
        self._txt_values.setReadOnly(state)
        self._txt_labels.setReadOnly(state)

    def isReadOnly(self):
        return _CalibrationWidget.isReadOnly(self) and \
            self._txt_values.isReadOnly() and \
            self._txt_labels.isReadOnly()

    def hasAcceptableInput(self):
        try:
            values = self._txt_values.text()
        except:
            return False

        try:
            labels = self._txt_labels.text()
        except:
            labels = None

        if labels is not None and len(values) != len(labels):
            return False

        return _CalibrationWidget.hasAcceptableInput(self) and \
            self._txt_values.hasAcceptableInput() and \
            self._txt_labels.hasAcceptableInput()

class CalibrationWidget(ParameterWidget):

    def __init__(self, parent=None):
        ParameterWidget.__init__(self, _Calibration, parent)

    def _init_ui(self):
        # Widgets
        self._combobox = QComboBox()
        self._stack = QStackedWidget()

        # Layouts
        layout = ParameterWidget._init_ui(self)
        layout.addRow(self._combobox)
        layout.addRow(self._stack)

        # Register classes
        self._widget_indexes = {}

        for entry_point in iter_entry_points('pyhmsa.gui.spec.condition.calibration'):
            widget_class = entry_point.load(require=False)
            widget = widget_class()
            self._combobox.addItem(widget.accessibleName().title())
            self._widget_indexes[widget.CLASS] = self._stack.addWidget(widget)
            widget.edited.connect(self.edited)

        # Signals
        self._combobox.currentIndexChanged.connect(self._on_combo_box)
        self._combobox.currentIndexChanged.connect(self.edited)

        return layout

    def _on_combo_box(self):
        # Old values
        oldwidget = self._stack.currentWidget()
        try:
            quantity = oldwidget._txt_quantity.text()
        except:
            quantity = None
        try:
            unit = oldwidget._txt_unit.text()
        except:
            unit = None

        # Change widget
        current_index = self._combobox.currentIndex()
        self._stack.setCurrentIndex(current_index)

        # Update values
        widget = self._stack.currentWidget()
        widget._txt_quantity.setText(quantity)
        widget._txt_unit.setText(unit)

    def parameter(self):
        return self._stack.currentWidget().calibration()

    def setParameter(self, calibration):
        index = self._widget_indexes[type(calibration)]
        self._combobox.setCurrentIndex(index)
        self._stack.setCurrentIndex(index)
        self._stack.currentWidget().setParameter(calibration)

    def calibration(self):
        return self.parameter()

    def setCalibration(self, calibration):
        self.setParameter(calibration)

    def setReadOnly(self, state):
        ParameterWidget.setReadOnly(self, state)
        self._combobox.setEnabled(not state)
        self._stack.currentWidget().setReadOnly(state)

    def isReadOnly(self):
        return ParameterWidget.isReadOnly(self) and \
            not self._combobox.isEnabled() and \
            self._stack.currentWidget().isReadOnly()

    def hasAcceptableInput(self):
        return ParameterWidget.hasAcceptableInput(self) and \
            self._stack.currentWidget().hasAcceptableInput()




"""
Custom widgets
"""

# Standard library modules.

# Third party modules.
from PySide.QtCore import QSize, Qt
from PySide.QtGui import \
    (QWidget, QPushButton, QVBoxLayout, QSizePolicy, QPainter, QBrush,
     QColorDialog)

# Local modules.

# Globals and constants variables.

class ColorDialogButton(QWidget):

    class _ColorButton(QPushButton):

        def __init__(self, *args, **kwargs):
            QPushButton.__init__(self, *args, **kwargs)

            # Variables
            self._color = Qt.red
            self._padding = 5

        def setColor(self, color):
            self._color = color

        def color(self):
            return self._color

        def setPadding(self, pad):
            self._padding = int(pad)

        def padding(self):
            return self._padding

        def paintEvent(self, event):
            QPushButton.paintEvent(self, event)

            color = self.color()
            padding = self.padding()

            rect = event.rect()
            painter = QPainter()
            painter.begin(self)
            painter.setBrush(QBrush(color))
            painter.setPen(Qt.NoPen)
            rect.adjust(padding, padding, -1 - padding, -1 - padding)
            painter.drawRect(rect)
            painter.end()

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)

        # Variables
        self._color = Qt.red

        # Widgets
        self._button = self._ColorButton()
        self._button.setFixedSize(QSize(25, 25))
        self._button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)

        # Layouts
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._button)
        self.setLayout(layout)

        # Signals
        self._button.clicked.connect(self._on_clicked)

        # Defaults
        self.setColor(Qt.red)

    def _on_clicked(self):
        dialog = QColorDialog(self.color())
        if not dialog.exec_():
            return
        self.setColor(dialog.selectedColor())

    def setColor(self, color):
        self._button.setColor(color)

    def color(self):
        return self._button.color()



"""
Specimen position condition widget
"""

# Standard library modules.
from operator import methodcaller

# Third party modules.
from PySide.QtGui import \
    (QHBoxLayout, QVBoxLayout, QLabel, QItemDelegate, QTableView,
     QToolBar, QMessageBox)
from PySide.QtCore import Qt, QAbstractTableModel, QModelIndex

# Local modules.
from pyhmsa.gui.util.parameter import \
    ParameterWidget, NumericalAttributeLineEdit
from pyhmsa.gui.util.icon import getIcon
from pyhmsa.gui.spec.condition.condition import _ConditionWidget

from pyhmsa.spec.condition.specimenposition import SpecimenPosition

# Globals and constants variables.

class SpecimenPositionWidget(_ConditionWidget):

    def __init__(self, inline=False, parent=None):
        self._inline = inline
        _ConditionWidget.__init__(self, SpecimenPosition, parent)

    def _init_ui(self):
        # Widgets
        self._txt_x = NumericalAttributeLineEdit(self._class.x)
        self._txt_y = NumericalAttributeLineEdit(self._class.y)
        self._txt_z = NumericalAttributeLineEdit(self._class.z)
        self._txt_r = NumericalAttributeLineEdit(self._class.r)
        self._txt_t = NumericalAttributeLineEdit(self._class.t)

        # Layouts
        if self._inline:
            layout = QVBoxLayout()

            layout_xyz = QHBoxLayout()
            layout_xyz.addWidget(QLabel('X'))
            layout_xyz.addWidget(self._txt_x)
            layout_xyz.addWidget(QLabel('Y'))
            layout_xyz.addWidget(self._txt_y)
            layout_xyz.addWidget(QLabel('Z'))
            layout_xyz.addWidget(self._txt_z)

            layout_rt = QHBoxLayout()
            layout_rt.addWidget(QLabel('R'))
            layout_rt.addWidget(self._txt_r)
            layout_rt.addWidget(QLabel('T'))
            layout_rt.addWidget(self._txt_t)

            layout.addLayout(layout_xyz)
            layout.addLayout(layout_rt)
        else:
            layout = _ConditionWidget._init_ui(self)
            layout.addRow('X', self._txt_x)
            layout.addRow('Y', self._txt_y)
            layout.addRow('Z', self._txt_z)
            layout.addRow('R', self._txt_r)
            layout.addRow('T', self._txt_t)

        # Signals
        self._txt_x.textEdited.connect(self.edited)
        self._txt_y.textEdited.connect(self.edited)
        self._txt_z.textEdited.connect(self.edited)
        self._txt_r.textEdited.connect(self.edited)
        self._txt_t.textEdited.connect(self.edited)

        return layout

    def parameter(self, parameter=None):
        parameter = _ConditionWidget.parameter(self, parameter)
        parameter.x = self._txt_x.text()
        parameter.y = self._txt_y.text()
        parameter.z = self._txt_z.text()
        parameter.r = self._txt_r.text()
        parameter.t = self._txt_t.text()
        return parameter

    def setParameter(self, condition):
        _ConditionWidget.setParameter(self, condition)
        self._txt_x.setText(condition.x)
        self._txt_y.setText(condition.y)
        self._txt_z.setText(condition.z)
        self._txt_r.setText(condition.r)
        self._txt_t.setText(condition.t)

class SpecimenPositionListWidget(ParameterWidget):

    class _SpecimenPositionModel(QAbstractTableModel):

        def __init__(self):
            QAbstractTableModel.__init__(self)
            self.positions = []

        def rowCount(self, *args, **kwargs):
            return len(self.positions)

        def columnCount(self, *args, **kwargs):
            return 5

        def data(self, index, role):
            if not index.isValid() or not (0 <= index.row() < len(self.positions)):
                return None
            if role != Qt.DisplayRole:
                return None

            position = self.positions[index.row()]
            column = index.column()
            if column == 0:
                return str(position.x) if position.x is not None else ''
            elif column == 1:
                return str(position.y) if position.y is not None else ''
            elif column == 2:
                return str(position.z) if position.z is not None else ''
            elif column == 3:
                return str(position.r) if position.r is not None else ''
            elif column == 4:
                return str(position.t) if position.t is not None else ''

        def headerData(self, section , orientation, role):
            if role != Qt.DisplayRole:
                return None
            if orientation == Qt.Horizontal:
                if section == 0:
                    return 'X'
                elif section == 1:
                    return 'Y'
                elif section == 2:
                    return 'Z'
                elif section == 3:
                    return 'R'
                elif section == 4:
                    return 'T'
            elif orientation == Qt.Vertical:
                return str(section + 1)

        def flags(self, index):
            if not index.isValid():
                return Qt.ItemIsEnabled

            return Qt.ItemFlags(QAbstractTableModel.flags(self, index) |
                                Qt.ItemIsEditable)

        def setData(self, index, value, role=Qt.EditRole):
            if not index.isValid() or \
                    not (0 <= index.row() < len(self.positions)):
                return False

            position = self.positions[index.row()]
            column = index.column()
            if column == 0:
                position.x = value
            elif column == 1:
                position.y = value
            elif column == 2:
                position.z = value
            elif column == 3:
                position.r = value
            elif column == 4:
                position.t = value

            return True

        def insertRows(self, row, count=1, parent=None):
            if count == 0:
                return False
            if parent is None:
                parent = QModelIndex()
            self.beginInsertRows(parent, row, row + count - 1)

            for i in range(count):
                self.positions.insert(row + i, SpecimenPosition())

            self.endInsertRows()
            return True

        def removeRows(self, row, count=1, parent=None):
            if count == 0:
                return False
            if parent is None:
                parent = QModelIndex()
            self.beginRemoveRows(parent, row, row + count - 1)

            self.positions = self.positions[:row] + self.positions[row + count:]

            self.endRemoveRows()
            return True

    class _SpecimenPositionDelegate(QItemDelegate):

        def __init__(self, parent=None):
            QItemDelegate.__init__(self, parent)

        def createEditor(self, parent, option, index):
            column = index.column()
            if column == 0:
                return NumericalAttributeLineEdit(SpecimenPosition.x, parent)
            elif column == 1:
                return NumericalAttributeLineEdit(SpecimenPosition.y, parent)
            elif column == 2:
                return NumericalAttributeLineEdit(SpecimenPosition.y, parent)
            elif column == 3:
                return NumericalAttributeLineEdit(SpecimenPosition.y, parent)
            elif column == 4:
                return NumericalAttributeLineEdit(SpecimenPosition.y, parent)
            else:
                return QItemDelegate.createEditor(self, parent, option, index)

        def setEditorData(self, editor, index):
            text = index.model().data(index, Qt.DisplayRole)
            column = index.column()
            if column == 0:
                editor.setText(text)
            elif column == 1:
                editor.setText(text)
            elif column == 2:
                editor.setText(text)
            elif column == 3:
                editor.setText(text)
            elif column == 4:
                editor.setText(text)
            else:
                QItemDelegate.setEditorData(self, editor, index)

        def setModelData(self, editor, model, index):
            column = index.column()
            if column == 0:
                model.setData(index, editor.text())
            elif column == 1:
                model.setData(index, editor.text())
            elif column == 2:
                model.setData(index, editor.text())
            elif column == 3:
                model.setData(index, editor.text())
            elif column == 4:
                model.setData(index, editor.text())
            else:
                return QItemDelegate.setModelData(self, editor, model, index)

    def __init__(self, parent=None):
        ParameterWidget.__init__(self, object, parent)

    def _init_ui(self):
        # Widgets
        self._table = QTableView()
        self._table.setModel(self._SpecimenPositionModel())
        self._table.setItemDelegate(self._SpecimenPositionDelegate(self))
        self._table.horizontalHeader().setStretchLastSection(True)

        self._toolbar = QToolBar()
        action_add = self._toolbar.addAction(getIcon("list-add"), "Add layer")
        action_remove = self._toolbar.addAction(getIcon("list-remove"), "Remove layer")

        # Layouts
        layout = ParameterWidget._init_ui(self)
        layout.addRow(self._table)
        layout.addRow(self._toolbar)

        # Signals
        action_add.triggered.connect(self._on_add)
        action_remove.triggered.connect(self._on_remove)

        return layout

    def _on_add(self):
        index = self._table.selectionModel().currentIndex()
        model = self._table.model()
        model.insertRows(index.row() + 1)

    def _on_remove(self):
        selection = self._table.selectionModel().selection().indexes()
        if len(selection) == 0:
            QMessageBox.warning(self, "Specimen position", "Select a position")
            return

        model = self._table.model()
        for row in sorted(map(methodcaller('row'), selection), reverse=True):
            model.removeRow(row)

    def parameter(self):
        positions = []
        for position in self._table.model().positions:
            positions.append(SpecimenPosition(position.x, position.y, position.z,
                                              position.r, position.t))
        return positions

    def setParameter(self, positions):
        model = self._table.model()
        model.positions = positions
        model.reset()

    def positions(self):
        return self.parameter()

    def setPositions(self, positions):
        self.setParameter(positions)

    def setReadOnly(self, state):
        ParameterWidget.setReadOnly(self, state)
        if state:
            trigger = QTableView.EditTrigger.NoEditTriggers
        else:
            trigger = QTableView.EditTrigger.AllEditTriggers
        self._table.setEditTriggers(trigger)
        self._toolbar.setEnabled(not state)

    def isReadOnly(self):
        return ParameterWidget.isReadOnly(self) and \
            self._table.editTriggers() == QTableView.EditTrigger.NoEditTriggers and \
            not self._toolbar.isEnabled()
